## -------------------------------------------------------------------------------------------------
## -- Project : MLPro - A Synoptic Framework for Standardized Machine Learning Tasks
## -- Package : mlpro
## -- Module  : howto_rl_001_types_of_reward.py
## -------------------------------------------------------------------------------------------------
## -- History :
## -- yyyy-mm-dd  Ver.      Auth.    Description
## -- 2021-05-30  0.0.0     DA       Creation
## -- 2021-05-31  1.0.0     DA       Release of first version
## -- 2021-09-11  1.0.1     MRD      Change Header information to match our new library name
## -------------------------------------------------------------------------------------------------

"""
Ver. 1.0.1 (2021-09-11)

This module shows how to create and interprete reward objects in own projects.
"""


from mlpro.bf.various import Log
from mlpro.rl.models import Reward



class MyLog(Log):
    C_TYPE      = 'Reward Demo'
    C_NAME      = ''


# 1 Some initial stuff
my_log = MyLog()

# 1.1 Unique agent ids
C_AGENT_1       = 1
C_AGENT_2       = 2
C_AGENT_3       = 3

# 1.2 Unique action ids
C_AGENT_1_ACT_1 = 1
C_AGENT_1_ACT_2 = 2
C_AGENT_1_ACT_3 = 3
C_AGENT_2_ACT_1 = 4


# 2 Rewards as single overall scalar values (independent from agents and actions)
my_log.log(Log.C_LOG_TYPE_I, 'Example for reward type C_TYPE_OVERALL:')
reward = Reward(p_type=Reward.C_TYPE_OVERALL)
reward.set_overall_reward(4.77)
my_log.log(Log.C_LOG_TYPE_I, 'Reward is just a scalar...', reward.get_agent_reward(0), '\n')


# 3 Rewards as scalar values for every agent
my_log.log(Log.C_LOG_TYPE_I, 'Example for reward type C_TYPE_EVERY_AGENT')
reward = Reward(p_type=Reward.C_TYPE_EVERY_AGENT)
my_log.log(Log.C_LOG_TYPE_I, 'Reward is a list with entries for each agent...')
reward.add_agent_reward(C_AGENT_1, 4.77)
my_log.log(Log.C_LOG_TYPE_I, 'Reward for agent 1 added:', reward.get_agent_reward(C_AGENT_1))
reward.add_agent_reward(C_AGENT_2, 5.19)
my_log.log(Log.C_LOG_TYPE_I, 'Reward for agent 2 added:', reward.get_agent_reward(C_AGENT_2))
reward.add_agent_reward(C_AGENT_3, 0.23)
my_log.log(Log.C_LOG_TYPE_I, 'Reward for agent 3 added:', reward.get_agent_reward(C_AGENT_3), '\n')

# 4 Rewards as scalar values for every agent and it's actions
my_log.log(Log.C_LOG_TYPE_I, 'Example for reward type C_TYPE_EVERY_ACTION')
reward = Reward(p_type=Reward.C_TYPE_EVERY_ACTION)
my_log.log(Log.C_LOG_TYPE_I, 'Reward is a list with entries for each agent and its action components...')
reward.add_action_reward(C_AGENT_1, C_AGENT_1_ACT_1, 1.23)
my_log.log(Log.C_LOG_TYPE_I, 'Reward for agent 1, action 1 added:', reward.get_action_reward(C_AGENT_1, C_AGENT_1_ACT_1))
reward.add_action_reward(C_AGENT_1, C_AGENT_1_ACT_2, 0.47)
my_log.log(Log.C_LOG_TYPE_I, 'Reward for agent 1, action 2 added:', reward.get_action_reward(C_AGENT_1, C_AGENT_1_ACT_2))
reward.add_action_reward(C_AGENT_1, C_AGENT_1_ACT_3, 1.63)
my_log.log(Log.C_LOG_TYPE_I, 'Reward for agent 1, action 3 added:', reward.get_action_reward(C_AGENT_1, C_AGENT_1_ACT_3))
reward.add_action_reward(C_AGENT_2, C_AGENT_2_ACT_1, 4.23)
my_log.log(Log.C_LOG_TYPE_I, 'Reward for agent 2, action 4 added:', reward.get_action_reward(C_AGENT_2, C_AGENT_2_ACT_1))



#! /usr/bin/env python
"""--------------------------------------------------------------------
COPYRIGHT 2015 Stanley Innovation Inc.

Software License Agreement:

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this
list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice,
this list of conditions and the following disclaimer in the documentation and/or
other materials provided with the distribution.

3. Neither the name of the copyright holder nor the names of its contributors
may be used to endorse or promote products derived from this software without
specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 
 file   robotiq_2f_gripper_driver.py

 brief  Driver for Robotiq 85 communication

 Platform: Linux/ROS Indigo
--------------------------------------------------------------------"""

"""
This Module defines the `Robotiq2FingerGripperDriver` and `Robotiq2FingerSimulatedGripperDriver` classes, which 
allow for the control and operation (also simulation of operation) of the 2 finger adaptive grippers from robotiq.

The end user should not need to use this class direcly since an instance of it is created with every action server
controlling a given gripper, and commanded by the user commands puubished by an action client instance.  
"""

from robotiq_2f_gripper_control.robotiq_2f_gripper import Robotiq2FingerGripper
from robotiq_2f_gripper_msgs.msg import RobotiqGripperCommand, RobotiqGripperStatus, CommandRobotiqGripperGoal
from sensor_msgs.msg import JointState
import numpy as np
import rospy
from enum import Enum

WATCHDOG_TIME = 1.0   # Max Time without communication with gripper allowed

class Robotiq2FingerGripperDriver:
    """
    This class represents an abstraction of a gripper driver, it handles the gripper connection, initialization,
    command send request, status request, and publishing of joint position to the `/joint_state` topic.   

    Args:
        comport: Name of the USB port to which the gripper is connected to.
        baud: Baudrate to use during Modbus protocol communication.
        stroke: Stroke of the gripper you are using, should be `0.085` or `0.140`.
        finger_joint: Name of the URDF joint to publish on the `/joint_state` topic. 

    Attributes:
        is_ready: Boolean indicating gripper is ready to take commands.
        _gripper: Instance of `robotiq_2f_gripper_control/Robotiq2FingerGripper` class representing the abstraction 
            of the real gripper.
        _gripper_joint_state_pub: Ros publisher for the `/joint_state` topic
        _driver_state: Internal machine state variable indicating the current gripper status
                            0: Driver not running
                            1: Driver is running
                            2: Gripper has been activated
    """
    def __init__(self, comport = '/dev/ttyACM0', baud = '115200', stroke = 0.085, joint_name='finger_joint'):
        self._comport = comport
        self._baud = baud
        self._joint_name = joint_name          
        

        # Instanciate and open communication with gripper.
        self._gripper = Robotiq2FingerGripper(device_id=0, stroke=stroke, comport=self._comport, baud=self._baud)
        
        self._max_joint_limit = 0.8
        if( self._gripper.stroke == 0.140 ):
            self._max_joint_limit = 0.7
        
        if not self._gripper.init_success:
            rospy.logerr("Unable to open commport to %s" % self._comport)
            return
        else:
            rospy.loginfo("Connection to gripper with stroke %.3f[m] on port %s successful" % ( self._gripper.stroke, self._comport))

        self._gripper_joint_state_pub = rospy.Publisher("/joint_states" , JointState, queue_size=10)        

        self._seq = 0
        self._prev_joint_pos = 0.0
        self._prev_joint_state_time = rospy.get_time() 
        self._driver_state = 0
        self.is_ready = False
        
        if not self._gripper.getStatus():
            rospy.logerr("Failed to contact gripper on port %s ... ABORTING" % self._comport)
            return                
                
        self._run_driver()
        self._last_update_time = rospy.get_time()
        
    def _clamp_position(self,cmd):
        out_of_bouds = False
        if (cmd <= 0.0):
            out_of_bouds = True
            cmd_corrected = 0.0
        elif (cmd > self._gripper.stroke):
            out_of_bouds = True
            cmd_corrected = self._gripper.stroke
        if(out_of_bouds):
            rospy.logdebug("Position (%.3f[m]) out of limits for %d[mm] gripper: \n- New position: %.3f[m]\n- Min position: %.3f[m]\n- Max position: %.3f[m]" % (cmd, self._gripper.stroke*1000, cmd_corrected, 0.0, self._gripper.stroke))
            cmd = cmd_corrected
        return cmd

    def _clamp_speed(self,cmd):
        out_of_bouds = False
        if (cmd <= 0.013):
            out_of_bouds = True
            cmd_corrected = 0.013
        elif (cmd > 0.101):
            out_of_bouds = True
            cmd_corrected = 0.1
        if(out_of_bouds):
            rospy.logdebug("Speed (%.3f[m/s]) out of limits for %d[mm] gripper: \n- New speed: %.3f[m/s]\n- Min speed: %.3f[m/s]\n- Max speed: %.3f[m/s]" % (cmd, self._gripper.stroke*1000, cmd_corrected, 0.013, 0.1))
            cmd = cmd_corrected
        return cmd
    
    def _clamp_force(self,cmd):
        out_of_bouds = False
        if (cmd < 0.0):
            out_of_bouds = True
            cmd_corrected = 0.0
        elif (cmd > 100.0):
            out_of_bouds = True
            cmd_corrected = 100.0
        if(out_of_bouds):
            # rospy.logdebug("Force (%.3f[%]) out of limits for %d[mm] gripper: \n- New force: %.3f[%]\n- Min force: %.3f[%]\n- Max force: %.3f[%]" % (cmd, self._gripper.stroke*1000, cmd_corrected, 0, 100))
            cmd = cmd_corrected
        return cmd
    
    def update_gripper_command(self,cmd):
        """
        Updates the driver internal goal/setpoint, which will be use to command the robotiq gripper/ 

        Args:
            cmd: Instance of `robotiq_2f_gripper_msgs/RobotiqGripperCommand` message, holding the most recent 
            user provided command parameters. See the message declaration for fields description 
        """
        if (True == cmd.emergency_release):
            self._gripper.activate_emergency_release(open_gripper=cmd.emergency_release_dir)
            return
        else:
            self._gripper.deactivate_emergency_release()

        if (True == cmd.stop):
            self._gripper.stop()
            
        else:
            pos = self._clamp_position(cmd.position)
            vel = self._clamp_speed(cmd.speed)
            force = self._clamp_force(cmd.force)
            # rospy.loginfo(" %.3f %.3f %.3f " % (pos,vel,force))
            self._gripper.goto(pos=pos,vel=vel,force=force)
            
    def get_current_gripper_status(self):
        """
        Public function to obtain the current gripper status.

        Returns:  Instance of `robotiq_2f_gripper_msgs/RobotiqGripperStatus` message. See the message declaration for fields description
        """
        status = RobotiqGripperStatus()
        status.header.stamp = rospy.get_rostime()
        status.header.seq = self._seq
        status.is_ready = self._gripper.is_ready()
        status.is_reset = self._gripper.is_reset()
        status.is_moving = self._gripper.is_moving()
        status.obj_detected = self._gripper.object_detected()
        status.fault_status = self._gripper.get_fault_status()
        status.position = self._gripper.get_pos()
        status.requested_position = self._gripper.get_req_pos()
        status.current = self._gripper.get_current()
        return status
        
    def _update_gripper_joint_state(self):
        """
        Private helper function to create a JointState message with the current gripper joint position
        """
        js = JointState()
        js.header.frame_id = ''
        js.header.stamp = rospy.get_rostime()
        js.header.seq = self._seq
        js.name = [self._joint_name]
        max_joint_limit = 0.8
        if( self._gripper.stroke == 0.140 ):
            max_joint_limit = 0.7
        pos = np.clip(max_joint_limit - ((max_joint_limit/self._gripper.stroke) * self._gripper.get_pos()), 0., max_joint_limit)
        js.position = [pos]
        dt = rospy.get_time() - self._prev_joint_state_time
        self._prev_joint_state_time = rospy.get_time()
        js.velocity = [(pos-self._prev_joint_pos)/dt]
        self._prev_joint_pos = pos
        return js
        
    def _run_driver(self):
        """
        Private function to test the comunication with the gripper and initialize it.
        WARNING: 
            Initialization of the gripper will generate a close motion of the real
            gripper. 

        Raises: 
            IOError: If Modbus RTU communication with the gripper is not achieved.
        """
        last_time = rospy.get_time()
        r = rospy.Rate(rospy.get_param('~rate', 100))
        
        while not rospy.is_shutdown() and self._driver_state != 2:
            # Check if communication is failing or taking too long
            dt = rospy.get_time() - last_time
            if (0 == self._driver_state):
                if (dt < 0.5):
                    self._gripper.deactivate_gripper()
                else:
                    self._driver_state = 1
            # If driver is not running, activate gripper 
            elif (1 == self._driver_state):
                is_gripper_activated = True
                self._gripper.activate_gripper()
                is_gripper_activated &= self._gripper.is_ready()
                if (is_gripper_activated):
                    rospy.loginfo("Gripper on port %s Activated" % self._comport)
                    self._driver_state = 2
                        
            success = True
            success &= self._gripper.sendCommand()
            success &= self._gripper.getStatus()
            if not success and not rospy.is_shutdown():
                rospy.logerr("Failed to initialize contact with gripper %d"% self._gripper.device_id)
            else:
                stat = RobotiqGripperStatus()
                #js = JointState()
                stat = self.get_current_gripper_status()
                js = self._update_gripper_joint_state()
                if stat.is_ready:
                    self.is_ready = True 
                # self._gripper_pub.publish(stat)
                self._gripper_joint_state_pub.publish(js)
                
            r.sleep()

    def update_driver(self):
        """
        Public function that:
            1. Sends the current driver command.
            2. Request the current gripper status.  
            3. Publish the gripper current joint position to the `/joint_state` topic.

        Raises:
            Log Error: If Modbus RTU communication with the gripper is not achieved.
        """
        # Try to update gripper command and status
        success = self._gripper.sendCommand()
        success &= self._gripper.getStatus()

        # Check if communication is broken
        update_time = rospy.get_time()
        if success:
            js = JointState()
            js = self._update_gripper_joint_state()
            self._gripper_joint_state_pub.publish(js)
            self._last_update_time = update_time
        
        # If communication failed, check if connection was truly lost
        elif (update_time - self._last_update_time) > WATCHDOG_TIME:
            rospy.logfatal("Failed to contact gripper on port: %s"% self._comport)
            # self._gripper.shutdown()
            # rospy.signal_shutdown("Communication to gripper lost")         
            

    def from_distance_to_radians(self, linear_pose ):
      """
      Private helper function to convert a command in meters to radians (joint value)
      """
      return np.clip(self._max_joint_limit - ((self._max_joint_limit/self._gripper.stroke) * linear_pose), 0.0, self._max_joint_limit)
  
    def from_radians_to_distance(self, joint_pose):
      """
      Private helper function to convert a joint position in radians to meters (distance between fingers)
      """
      return np.clip(self._gripper.stroke - ((self._gripper.stroke/self._max_joint_limit) * joint_pose), 0.0, self._max_joint_limit)
      
    def get_current_joint_position(self):
      return self.from_distance_to_radians(self._gripper.get_pos())
    ######################################################################################################
    # STATIC functions for fast control of the gripper.

    @staticmethod
    def goto( client, pos, speed=0.1, force=120, block = True ):
        """
        Static function to update the gripper command

        Args:
            client: `SimpleActionClient` instance connected to the action server holding a robotiq gripper
            instance.
            pos: [m] Position (distance in between fingers) in meters desired for the gripper.
            speed: [m/s] Motion speed in meters over seconds. Min value: 0.013[m/s] - Max value: 0.1[m/s]
            force: [%] Force value to apply in gripper motion, see robotiq manuals to calculate an appropiate
                    gripping force value in Newtons.  
            block: Boolean indicating whether to lock the current thread until command has been reached or not.
        """
        goal = CommandRobotiqGripperGoal()
        goal.emergency_release = False
        goal.stop = False
        goal.position = pos
        goal.speed = speed
        goal.force = force

        # Sends the goal to the gripper.
        if block:   
            client.send_goal(goal)
            client.wait_for_result()
        else:
            client.send_goal(goal)
    
    @staticmethod
    def close( client, speed=0.1, force=120, block = True,):
        """
        Static function to close the gripper 

        Args:
            client: `SimpleActionClient` instance connected to the action server holding a robotiq gripper
            instance.
            speed: [m/s] Motion speed in meters over seconds. Min value: 0.013[m/s] - Max value: 0.1[m/s]
            force: [%] Force value to apply in gripper motion, see robotiq manuals to calculate an appropiate
                    gripping force value in Newtons.  
            block: Boolean indicating whether to lock the current thread until command has been reached or not.
        """
        goal = CommandRobotiqGripperGoal()
        goal.emergency_release = False
        goal.stop = False
        goal.position = 0.0
        goal.speed = speed
        goal.force = force

        # Sends the goal to the gripper.
        if block:
            client.send_goal_and_wait(goal)
        else:
            client.send_goal(goal)

    @staticmethod
    def open( client, speed=0.1, force=120, block = True):
        """
        Static function to open the gripper 

        Args:
            client: `SimpleActionClient` instance connected to the action server holding a robotiq gripper
            instance.
            speed: [m/s] Motion speed in meters over seconds. Min value: 0.013[m/s] - Max value: 0.1[m/s]
            force: [%] Force value to apply in gripper motion, see robotiq manuals to calculate an appropiate
                    gripping force value in Newtons.  
            block: Boolean indicating whether to lock the current thread until command has been reached or not.
        """
        goal = CommandRobotiqGripperGoal()
        goal.emergency_release = False
        goal.stop = False
        goal.position = 255 # Use max value to make it stroke independent
        goal.speed = speed
        goal.force = force

        # Sends the goal to the gripper.
        if block:
            client.send_goal_and_wait(goal)
        else:
            client.send_goal(goal)

    @staticmethod
    def stop( client, block = True):
        """
        Static function to stop gripper motion

        Args:
            client: `SimpleActionClient` instance connected to the action server holding a robotiq gripper
            instance.
            block: Boolean indicating whether to lock the current thread until command has been reached or not.
        """
        goal = CommandRobotiqGripperGoal()
        goal.emergency_release = False
        goal.stop = False

        # Sends the goal to the gripper.
        if block:
            client.send_goal_and_wait(goal)
        else:
            client.send_goal(goal)

    @staticmethod
    def emergency_release( client ):
        """
        Static function to trigger an emergency release motion.

        Args:
            client: `SimpleActionClient` instance connected to the action server holding a robotiq gripper
            instance.
        """
        goal = CommandRobotiqGripperGoal()
        goal.emergency_release = True
        client.send_goal_and_wait(goal)

class Robotiq2FingerSimulatedGripperDriver:
    """
    This class represents an abstraction of a gripper driver for a SIMULATED gripper,
    by simulating the operation/response of the gripper to a given command and the publishing 
    of the simulated gripper joint position to the `/joint_state` topic.  

    Args:
        stroke: Stroke of the gripper you are using, should be `0.085` or `0.140`.
        finger_joint: Name of the URDF joint to publish on the `/joint_state` topic. 

    Attributes:
        is_ready: Boolean indicating gripper is ready to take commands.
        _gripper_joint_state_pub: Ros publisher for the `/joint_state` topic.
        _current_joint_pos: [radians] Position of the simulated joint in radians.
        _current_goal: Instance of `RobotiqGripperCommand` message holding the latest user command.
    """
    def __init__(self, stroke=0.085, joint_name='finger_joint'):
        self._stroke = stroke
        self._joint_name = joint_name
        self._current_joint_pos = 0.0                                 
        self._prev_time = rospy.get_time()
        self._current_goal = CommandRobotiqGripperGoal()
        self._current_goal.position = self._stroke
        self._gripper_joint_state_pub = rospy.Publisher("/joint_states" , JointState, queue_size=10)  
        self.is_ready = True
        self._is_moving = False
        self._max_joint_limit = 0.8
        if( self._stroke == 0.140 ):
            self._max_joint_limit = 0.7


    def update_driver(self):
        """
        Public function simulating the gripper state change given the current gripper command and the time
        difference between the last call to this function.
        """
        delta_time = rospy.get_time() - self._prev_time
        linear_position_goal = self._current_goal.position                                                   #[mm]
        joint_goal = self.from_distance_to_radians(linear_position_goal)                                     #[rad]
        position_increase = (delta_time * self._current_goal.speed) * (self._max_joint_limit/self._stroke) 
        if( abs(joint_goal - self._current_joint_pos) > position_increase ):
            self._current_joint_pos += position_increase if (joint_goal - self._current_joint_pos) > 0 else -position_increase
            self._is_moving = True
        else:
            self._current_joint_pos = joint_goal
            self._is_moving = False
        js = self._update_gripper_joint_state()
        self._gripper_joint_state_pub.publish(js)
        self._prev_time = rospy.get_time()


    def update_gripper_command(self, goal_command):
        """
        Updates the driver internal goal/setpoint, which will be use to command the robotiq gripper/ 

        Args:
            goal_command: Instance of `robotiq_2f_gripper_msgs/RobotiqGripperCommand` message, holding the most recent 
            user provided command parameters. See the message declaration for fields description 
        """
        self._current_goal = goal_command
        self._current_goal.position = self._clamp_position(goal_command.position)
        self._current_goal.speed = self._clamp_speed(goal_command.speed)

    def get_current_gripper_status(self):
        """
        Public function to obtain the current gripper status.

        Returns:  Instance of `robotiq_2f_gripper_msgs/RobotiqGripperStatus` message. See the message declaration for fields description
        """
        status = RobotiqGripperStatus()
        status.header.stamp = rospy.get_rostime()
        status.header.seq = 0
        status.is_ready = True
        status.is_reset = False
        status.is_moving = self._is_moving
        status.obj_detected = False
        status.fault_status = False
        status.position = self.from_radians_to_distance(self._current_joint_pos) #[mm]
        status.requested_position = self._current_goal.position                                                                                 #[mm]
        status.current = 0.0
        return status

    def get_current_joint_position(self):
      return self._current_joint_pos

    def from_distance_to_radians(self, linear_pose ):
      """
      Private helper function to convert a command in meters to radians (joint value)
      """
      return np.clip(self._max_joint_limit - ((self._max_joint_limit/self._stroke) * linear_pose), 0.0, self._max_joint_limit)
  
    def from_radians_to_distance(self, joint_pose):
      """
      Private helper function to convert a joint position in radians to meters (distance between fingers)
      """
      return np.clip(self._stroke - ((self._stroke/self._max_joint_limit) * joint_pose), 0.0, self._max_joint_limit)

    def _update_gripper_joint_state(self):
        """
        Private helper function to create a JointState message with the current gripper joint position
        """
        js = JointState()
        js.header.frame_id = ''
        js.header.stamp = rospy.get_rostime()
        js.header.seq = 0
        js.name = [self._joint_name]
        pos = np.clip( self._current_joint_pos, 0.0, self._max_joint_limit)
        js.position = [pos]
        js.velocity = [self._current_goal.speed]
        self._prev_joint_pos = pos
        return js

    def _clamp_position(self,cmd):
        if (cmd <= 0.0):
            cmd = 0.0
        elif (cmd >= self._stroke):
            cmd = self._stroke
        return cmd
    
    def _clamp_speed(self,cmd):
        if (cmd < 0.0):
            cmd = 0.01
        elif (cmd > 0.101):
            cmd = 0.1
        return cmd




## -------------------------------------------------------------------------------------------------
## -- Project : MLPro - A Synoptic Framework for Standardized Machine Learning Tasks
## -- Package : mlpro.gt
## -- Module  : models.py
## -------------------------------------------------------------------------------------------------
## -- History :
## -- yyyy-mm-dd  Ver.      Auth.    Description
## -- 2021-05-21  0.0.0     DA       Creation
## -- 2021-06-06  1.0.0     DA       Release of first version
## -- 2021-06-25  1.0.1     DA       Class GameBoard: introduction of internal constant C_REWARD_TYPE
## -- 2021-08-28  1.0.2     DA       Adjustments after changings on rl models
## -- 2021-09-11  1.0.3     MRD      Change Header information to match our new library name
## -- 2021-11-16  1.1.0     DA       Refactoring
## -- 2021-12-03  1.1.1     DA       Refactoring
## -- 2021-12-07  1.1.2     DA       Refactoring
## -- 2021-12-09  1.1.3     DA       Class GTTraining: refactoring
## -------------------------------------------------------------------------------------------------

"""
Ver. 1.1.2 (2021-12-07)

This module provides model classes for tasks related to cooperative Game Theory.
"""

from datetime import timedelta
from mlpro.bf.various import Log
from mlpro.rl.models import *


## -------------------------------------------------------------------------------------------------
## -------------------------------------------------------------------------------------------------
class GameBoard(Environment):
    """
    Model class for a game theoretical game board. See super class for more information.
    """

    C_TYPE = 'Game Board'
    C_REWARD_TYPE = Reward.C_TYPE_EVERY_AGENT

    ## -------------------------------------------------------------------------------------------------
    def _compute_reward(self, p_state_old: State, p_state_new: State) -> Reward:
        reward = Reward(self.get_reward_type())

        for player_id in self._last_action.get_agent_ids():
            reward.add_agent_reward(player_id, self._utility_fct(p_state_new, player_id))

        return reward

    ## -------------------------------------------------------------------------------------------------
    def _utility_fct(self, p_state: State, p_player_id):
        """
        Computes utility of given player. To be redefined.
        """

        raise NotImplementedError


## -------------------------------------------------------------------------------------------------
## -------------------------------------------------------------------------------------------------
class PGameBoard(GameBoard):
    """
    Model class for a potential game theoretical game board. See super class for more information.
    """

    C_TYPE = 'Potential Game Board'

    ## -------------------------------------------------------------------------------------------------
    def compute_potential(self):
        """
        Computes (weighted) potential level of the game board.
        """

        if self._last_action == None: return 0
        self.potential = 0

        for player_id in self._last_action.get_agent_ids():
            self.potential = self.potential + (
                        self._utility_fct(player_id) * self._last_action.get_elem(player_id).get_weight())

        return self.potential


## -------------------------------------------------------------------------------------------------
## -------------------------------------------------------------------------------------------------
class Player(Agent):
    """
    This class implements a game theoretical player model. See super class for more information.
    """

    C_TYPE = 'Player'


## -------------------------------------------------------------------------------------------------
## -------------------------------------------------------------------------------------------------
class MultiPlayer(MultiAgent):
    """
    This class implements a game theoretical model for a team of players. See super class for more 
    information.
    """

    C_TYPE = 'Multi-Player'

    ## -------------------------------------------------------------------------------------------------
    def add_player(self, p_player: Player, p_weight=1.0) -> None:
        super().add_agent(p_agent=p_player, p_weight=p_weight)


## -------------------------------------------------------------------------------------------------
## -------------------------------------------------------------------------------------------------
class Game(RLScenario):
    """
    This class implements a game consisting of a game board and a (multi-)player. See super class for 
    more information.
    """

    C_TYPE = 'Game'


## -------------------------------------------------------------------------------------------------
## -------------------------------------------------------------------------------------------------
class GTTraining(RLTraining):
    """
    This class implements a standardized episodical training process. See super class for more 
    information.

    Parameters
    ----------
    p_game_cls 
        Name of GT game class, compatible to/inherited from class Game.
    p_cycle_limit : int
        Maximum number of training cycles (0=no limit). Default = 0.
    p_cycles_per_epi_limit : int
        Optional limit of cycles per episode (0=no limit, -1=get environment limit). Default = -1.    
    p_adaptation_limit : int
        Maximum number of adaptations (0=no limit). Default = 0.
    p_stagnation_limit : int
        Optional limit of consecutive evaluations without training progress. Default = 0.
    p_eval_frequency : int
        Optional evaluation frequency (0=no evaluation). Default = 0.
    p_eval_grp_size : int
        Number of evaluation episodes (eval group). Default = 0.
    p_hpt : HyperParamTuner
        Optional hyperparameter tuner (see class mlpro.bf.ml.HyperParamTuner). Default = None.
    p_hpt_trials : int
        Optional number of hyperparameter tuning trials. Default = 0. Must be > 0 if p_hpt is supplied.
    p_path : str
        Optional destination path to store training data. Default = None.
    p_collect_states : bool
        If True, the environment states will be collected. Default = True.
    p_collect_actions : bool
        If True, the agent actions will be collected. Default = True.
    p_collect_rewards : bool
        If True, the environment reward will be collected. Default = True.
    p_collect_training : bool
        If True, global training data will be collected. Default = True.
    p_visualize : bool
        Boolean switch for env/agent visualisation. Default = False.
    p_logging
        Log level (see constants of class mlpro.bf.various.Log). Default = Log.C_LOG_WE.

    """

    C_NAME = 'GT'

    ## -------------------------------------------------------------------------------------------------
    def __init__(self, **p_kwargs):
        kwargs = p_kwargs.copy()
        kwargs['p_scenario_cls'] = kwargs['p_game_cls']
        kwargs.pop('p_game_cls')
        super().__init__(**kwargs)



#!/usr/bin/env python
from click import launch
import gym
from .task_envs.task_envs_list import RegisterOpenAI_Ros_Env
import roslaunch
import rospy
import rospkg
import os
import sys
import subprocess


def StartOpenAI_ROS_Environment(task_and_robot_environment_name, max_episode_steps=10000):
    """
    It Does all the stuff that the user would have to do to make it simpler
    for the user.
    This means:
    0) Registers the TaskEnvironment wanted, if it exists in the Task_Envs.
    2) Checks that the workspace of the user has all that is needed for launching this.
    Which means that it will check that the robot spawn launch is there and the worls spawn is there.
    4) Launches the world launch and the robot spawn.
    5) It will import the Gym Env and Make it.
    """
    rospy.logwarn("Env: {} will be imported".format(
        task_and_robot_environment_name))
    result = RegisterOpenAI_Ros_Env(task_env=task_and_robot_environment_name,
                                    max_episode_steps=max_episode_steps)

    if result:
        rospy.logwarn("Register of Task Env went OK, lets make the env..."+str(task_and_robot_environment_name))
        env = gym.make(task_and_robot_environment_name)
    else:
        rospy.logwarn("Something Went wrong in the register")
        env = None

    return env


class ROSLauncher(object):
    def __init__(self, rospackage_name, launch_file_name, launch_arguments={}, use_popen=False, ros_ws_abspath="/home/user/simulation_ws"):

        self._rospackage_name = rospackage_name
        self._launch_file_name = launch_file_name
        self._launch_arguments = launch_arguments

        self.rospack = rospkg.RosPack()

        # Check Package Exists
        try:
            pkg_path = self.rospack.get_path(rospackage_name)
            rospy.logdebug("Package FOUND...")
        except rospkg.common.ResourceNotFound:
            rospy.logwarn("Package NOT FOUND...")

        # Now we check that the Package path is inside the ros_ws_abspath
        # This is to force the system to have the packages in that ws, and not in another.
        if ros_ws_abspath in pkg_path:
            rospy.logdebug("Package FOUND in the correct WS!")
        else:
            rospy.logwarn("Package FOUND in "+pkg_path +
                          ", BUT not in the ws="+ros_ws_abspath+"...")

        # If the package was found then we launch
        if pkg_path:
            rospy.loginfo(
                ">>>>>>>>>>Package found in workspace-->"+str(pkg_path))
            launch_dir = os.path.join(pkg_path, "launch")
            path_launch_file_name = os.path.join(launch_dir, launch_file_name)

            # Convert Launch argument from dictionary to proper argument
            launch_args = []
            for key, val in self._launch_arguments.items():
                launch_args.append(str(key)+":="+str(val))
        

            rospy.logwarn("path_launch_file_name=="+str(path_launch_file_name))
            if use_popen:
                source_env_command = "source "+ros_ws_abspath+"/devel/setup.bash"
                roslaunch_command = "roslaunch  {0} {1} {2}".format(rospackage_name, launch_file_name, launch_args)
                command = source_env_command+roslaunch_command
                rospy.logwarn("Launching command="+str(command))

                p = subprocess.Popen(['/bin/bash', '-c', source_env_command, "&&", roslaunch_command], shell=True)
                
            else:
                self.uuid = roslaunch.rlutil.get_or_generate_uuid(None, False)
                roslaunch.configure_logging(self.uuid)
                self.launch = roslaunch.parent.ROSLaunchParent(
                    self.uuid, [(path_launch_file_name, launch_args)])
                self.launch.start()
            


            rospy.loginfo(">>>>>>>>>STARTED Roslaunch-->" +
                          str(self._launch_file_name))
        else:
            assert False, "No Package Path was found for ROS apckage ==>" + \
                str(rospackage_name)



#!/usr/bin/env python
import rospy
import rospkg
import roslaunch
import numpy as np
import sys
import subprocess

import moveit_commander
from moveit_msgs.msg import PlanningScene
from geometry_msgs.msg import Pose
from std_msgs.msg import String, Bool
from sensor_msgs.msg import PointCloud2, JointState
from actionlib_msgs.msg import GoalStatusArray
import sensor_msgs.point_cloud2 as pc2
from ctypes import cast,POINTER,pointer,c_float,c_uint32
from gazebo_msgs.srv import GetWorldProperties, GetModelState
from gazebo_msgs.msg import ContactsState
from ur_dashboard_msgs.srv import IsProgramRunning

from openai_ros import robot_gazebo_env
from openai_ros import robot_real_env
from openai_ros.openai_ros_common import ROSLauncher


class UR5LabRealEnv(robot_real_env.RobotRealEnv):
    """Superclass for all UR5LabEnv environments.
    """
    

    def __init__(self, ros_ws_abspath):
        """
        Initializes a new UR5LabEnv environment.

        To check any topic we need to have the simulations running, we need to do two things:
        1) Make sure simulation is unpaused: without that, the stream of data doesnt flow.
        2) If the simulation was already running for some reason, reset the controlers.
        This has to do with the fact that some plugins with tf, dont understand the reset of the simulation
        and need to be reseted to work properly.

        Actuators Topic List:
        * /gripper_controller/command
        
        Additional Data:
        * moveit_commander
        * gripper contact sensor
        """
        rospy.loginfo("Start UR5LabEnv INIT...")
        
        # We launch the ROSlaunch that spawns the robot into the world
        visualize = rospy.get_param("visualize", False)
        sim_mode = rospy.get_param("sim", True)
        robot_ip = rospy.get_param("robot_ip", "")
        reverse_ip = rospy.get_param("reverse_ip", "")
        reverse_port = rospy.get_param("reverse_port", 50001)

        if not sim_mode:
            assert robot_ip != "", "Please set the Robot IP"

        # Start the ur_robot_driver
        ROSLauncher(rospackage_name="ur_gazebo",
                    launch_file_name="ur5_lab.launch",
                    launch_arguments=dict(gui=visualize, sim=sim_mode, robot_ip=robot_ip, 
                    reverse_ip=reverse_ip, reverse_port=reverse_port),
                    ros_ws_abspath=ros_ws_abspath)
        
        # # Wait the service
        rospy.loginfo("Wait for the robot")
        rospy.wait_for_service('/ur_hardware_interface/dashboard/program_running')

        # Call the service
        robot_program_running_call = rospy.ServiceProxy('/ur_hardware_interface/dashboard/program_running', IsProgramRunning)
        self.program_robot_running = False
        
        # Keep calling until is True
        while self.program_robot_running is False and not rospy.is_shutdown():
            try:
                rospy.loginfo("Wait for the robot")
                rospy.sleep(1)
                self.program_robot_running = robot_program_running_call().program_running
            except rospy.ServiceException as e:
                rospy.logerr("Cannot retrieve robot program status")
                
        # Internal Vars
        self.controllers_list = []
        
        # It doesnt use namespace
        self.robot_name_space = ""

        # We launch the init function of the Parent Class robot_gazebo_env.RobotGazeboEnv
        super(UR5LabRealEnv, self).__init__(controllers_list=self.controllers_list,
                                            robot_name_space=self.robot_name_space,
                                            reset_controls=False)

        # self._gripper = rospy.Publisher('/gripper_controller/command', String, 
        #                                    queue_size=1)
    
        # Subscribed to robot /ur_hardware_interface/robot_program_running
        # Wait until running mode
        # /ur_hardware_interface/dashboard/program_running
        rospy.Subscriber("/ur_hardware_interface/robot_program_running", Bool, self._connection_state_callback)

        self._check_all_systems_ready()

        rospy.Subscriber("/joint_states", JointState, self._joints_state_callback)
        rospy.Subscriber("/move_group/status", GoalStatusArray, self._plan_status_feedback)
        self._tcp_pose_pub = rospy.Publisher("/tcp_pose", Pose, queue_size=10)
        self.last_contact_r = 0
        self.last_contact_l = 0
        

        self.uuid = roslaunch.rlutil.get_or_generate_uuid(None, False)
        roslaunch.configure_logging(self.uuid)
        ros_pack = rospkg.RosPack()
        ros_pack = ros_pack.get_path('ur5_moveit_config')
        ros_pack += '/launch/move_group.launch'
        self.launch = roslaunch.parent.ROSLaunchParent(self.uuid, [ros_pack])
        self.launch.start()

        # source_env_command = "source "+ros_ws_abspath+"/devel/setup.bash;"
        # roslaunch_command = "roslaunch  {0} {1}".format("ur5_moveit_config", "move_group.launch")
        # command = source_env_command+roslaunch_command
        # rospy.logwarn("Launching command="+str(command))

        # p = subprocess.Popen(command, shell=True)
        rospy.loginfo("move_group started")
        
        self._move = MoveIt()
        
        rospy.logdebug("UR5LabEnv pausing")
        rospy.logdebug("Finished UR5LabEnv INIT...")
        
    # Methods needed by the RobotGazeboEnv
    # ----------------------------
    def close(self):
        self.launch.launch.shutdown()
        super().close()
        
    def _check_all_systems_ready(self):
        """
        Checks that all the sensors, publishers and other simulation systems are
        operational.
        """
        rospy.logdebug("UR5LabEnv check_all_systems_ready...")
        self._check_all_publishers_ready()
        self._check_all_sensors_ready()
        rospy.logdebug("END UR5LabEnv _check_all_systems_ready...")
        return True

    def _check_all_publishers_ready(self):
        rospy.logdebug("START ALL PUBLISHER")
        # self._check_gripper_pub_ready()
        rospy.logdebug("All Publishers READY")
        
    def _check_gripper_pub_ready(self):
        rate = rospy.Rate(10)  # 10hz
        while (self._gripper.get_num_connections() == 0 and not rospy.is_shutdown()):
            rospy.logdebug(
                "No susbribers to _gripper yet so we wait and try again")
            try:
                rate.sleep()
            except rospy.ROSInterruptException:
                # This is to avoid error when world is rested, time when backwards.
                pass
        rospy.logdebug("_gripper Publisher Connected")
        
    def _check_all_sensors_ready(self):
        rospy.logdebug("START ALL SENSORS READY")
        self._check_joint_states_ready()
        rospy.logdebug("ALL SENSORS READY")

    def _JointState_numpy(self, data):
        data.name = list(data.name)
        data.position = list(data.position)
        manipulator_index = [data.name.index('shoulder_pan_joint'), data.name.index('shoulder_lift_joint'),
                                data.name.index('elbow_joint'), data.name.index('wrist_1_joint'),
                                data.name.index('wrist_2_joint'), data.name.index('wrist_3_joint')]
                                
        manipulator_joints_position = [data.position[i] for i in manipulator_index]

        return manipulator_joints_position
        
    def _check_joint_states_ready(self): #check published joint names and rearrange them for the move joint_goal
        self.joint_states = None
        self.gripper_states = None
        rospy.logdebug("Waiting for /joint_states to be READY...")
        while self.joint_states is None and not rospy.is_shutdown():
            try:
                self.joint_states = self._JointState_numpy(rospy.wait_for_message("/joint_states", JointState, timeout=2.0))
                rospy.sleep(1)
                rospy.logdebug("Current /joint_states READY=>")

            except:
                rospy.logerr("Current /joint_states not ready yet, retrying for getting joint_states")
        return self.joint_states

    def _check_no_motion_plan_ready(self):
        self.no_motion_plan = None
        rospy.logdebug("Waiting for Goal Status to be READY...")
        while self.no_motion_plan is None and not rospy.is_shutdown():
            try:
                self.no_motion_plan = self._no_motion_plan(rospy.wait_for_message("/move_group/status", GoalStatusArray, timeout=2.0))
                rospy.logdebug("Current Goal Status READY=>")

            except:
                rospy.logerr("Current Goal Status not ready yet, retrying for getting joint_states")
        return self.no_motion_plan

    def _no_motion_plan(self, data):
        if data.status_list:
            if data.status_list[-1].text in "No motion plan found. No execution attempted.":
                return True
        return False
        
    def _joints_state_callback(self, data):
        self.joint_states = self._JointState_numpy(data)
        return 

    def _connection_state_callback(self, data):
        self.robot_connected = data.data
        return

    def _plan_status_feedback(self, data):
        self.no_motion_plan = self._no_motion_plan(data)
        return

        
    # Methods that the TrainingEnvironment will need to define here as virtual
    # because they will be used in RobotGazeboEnv GrandParentClass and defined in the
    # TrainingEnvironment.
    # ----------------------------
    def _set_init_pose(self):
        """
        Sets the Robot in its init pose
        """
        raise NotImplementedError()

    def _init_env_variables(self):
        """
        Inits variables needed to be initialised each time 
        we reset at the start of an episode.
        """
        raise NotImplementedError()

    def _compute_reward(self, observations, done):
        """
        Calculates the reward to give based on the observations given.
        """
        raise NotImplementedError()

    def _set_action(self, action):
        """
        Applies the given action to the simulation.
        """
        raise NotImplementedError()

    def _get_obs(self):
        raise NotImplementedError()

    def _is_done(self, observations):
        """
        Checks if episode done based on observations given.
        """
        raise NotImplementedError()

    # Methods that the TrainingEnvironment will need.
    # ----------------------------
    def open_hand(self):
        """
        When called it opens robots hand
        """
        # self._gripper.publish("open")
        pass
       
    def close_hand(self):
        """
        When called it closes robots hand
        """
        # self._gripper.publish("close")
        pass
        
    def get_joint_states(self): 
        joint = self.joint_states
        return joint 
        
    def get_gripper_states(self):
        joint = self.gripper_states
        return joint 
    
    def get_pc(self):
        return np.asarray(self.pc.points)
    
    def get_ee_pose(self): 
        """
        Returns the endeffector pose
        We make sure the simulation is unpaused because this calss is a service call.
        """
        rospy.logdebug("START getting endeffector pose")
        gripper_pose = self._move.ee_pose()
        rospy.logdebug("END getting endeffector pose")
        return gripper_pose   
        
    def get_ee_rpy(self): 
        rospy.logdebug("START getting endeffector rpy")
        gripper_rpy = self._move.ee_rpy()
        rospy.logdebug("END getting endeffector rpy")
        return gripper_rpy
        
    def move_ee(self, desired_pose): 
        """
        Moves the endeffector to the pose given,
        relative pose to world frame
        Currently it is only accepting position
        """
        pose_goal = Pose()
        pose_goal.position.x = desired_pose[0]
        pose_goal.position.y = desired_pose[1]
        pose_goal.position.z = desired_pose[2]
        pose_goal.orientation.x = desired_pose[3]
        pose_goal.orientation.y = desired_pose[4]
        pose_goal.orientation.z = desired_pose[5]
        pose_goal.orientation.w = desired_pose[6]
        
        
        result = self._move.ee_traj(pose_goal)
        return result
           
    def move_joints(self, joint_goal, wait=True): 
        """
        Moves the joints of the robot directly
        """
        result = self._move.joint_traj(joint_goal, wait)
        return result
    
    def check_pose(self, pose):
        pos = Pose()
        pos.position.x = pose[0]
        pos.position.y = pose[1]
        pos.position.z = pose[2]
        pos.orientation.x = pose[3]
        pos.orientation.y = pose[4]
        pos.orientation.z = pose[5]
        pos.orientation.w = pose[6]
        
        return self._move.check_pos(pos)[0]
    
    def get_observation(self):
        obs = self._get_obs()
        return obs

class MoveIt(object):
    def __init__(self):
        rospy.logdebug("===== MoveIt")
        moveit_commander.roscpp_initialize(sys.argv)
        self.robot = moveit_commander.RobotCommander()
        self.scene = moveit_commander.PlanningSceneInterface()
        self.move_group = moveit_commander.MoveGroupCommander("manipulator")
        self.move_group.set_num_planning_attempts(1)
        self.move_group.allow_replanning(False)
        self.eef_link = self.move_group.get_end_effector_link()
        self.joint_index()
        
        rospy.logdebug("============")

    def ee_traj(self, goal):
        self.move_group.clear_pose_targets()
        self.move_group.set_pose_target(goal)
        plan = self.move_group.go(wait=True)
        self.move_group.stop()
        self.move_group.clear_pose_targets()
        return plan

    def joint_traj(self, positions_array, wait=True):
        self.move_group.clear_pose_targets()   
        positions_array = [positions_array[i] for i in self.joint_name_index]
        result = self.move_group.go(positions_array, wait=wait)
        self.move_group.stop()
        return result
        
    def robot_joints(self):
        return self.move_group.get_current_joint_values()
        
    def check_pos(self, pose):
        return self.move_group.plan(pose)
    
    def joint_index(self):
        active_joints = self.move_group.get_active_joints()
        self.joint_name_index = [active_joints.index('shoulder_pan_joint'), active_joints.index('shoulder_lift_joint'),
                                active_joints.index('elbow_joint'), active_joints.index('wrist_1_joint'),
                                active_joints.index('wrist_2_joint'), active_joints.index('wrist_3_joint')]
        return
        
    def ee_pose(self):
        gripper_pose = self.move_group.get_current_pose().pose
        return gripper_pose

    def ee_rpy(self, request):
        gripper_rpy = self.move_group.get_current_rpy()
        return gripper_rpy
            



class UR5LabSimEnv(robot_gazebo_env.RobotGazeboEnv):
    """Superclass for all UR5LabEnv environments.
    """
    

    def __init__(self, ros_ws_abspath):
        """
        Initializes a new UR5LabEnv environment.

        To check any topic we need to have the simulations running, we need to do two things:
        1) Make sure simulation is unpaused: without that, the stream of data doesnt flow.
        2) If the simulation was already running for some reason, reset the controlers.
        This has to do with the fact that some plugins with tf, dont understand the reset of the simulation
        and need to be reseted to work properly.

        Actuators Topic List:
        * /gripper_controller/command
        
        Additional Data:
        * moveit_commander
        * gripper contact sensor
        """
        rospy.loginfo("Start UR5LabEnv INIT...")
        
        # We launch the ROSlaunch that spawns the robot into the world
        visualize = rospy.get_param("visualize", False)
        sim_mode = rospy.get_param("sim", True)
        robot_ip = rospy.get_param("robot_ip", "")
        reverse_ip = rospy.get_param("reverse_ip", "")
        reverse_port = rospy.get_param("reverse_port", 50001)

        if not sim_mode:
            assert robot_ip != "", "Please set the Robot IP"

        self.launch = ROSLauncher(rospackage_name="ur_gazebo",
                    launch_file_name="ur5_lab.launch",
                    launch_arguments=dict(gui=visualize, sim=sim_mode, robot_ip=robot_ip, 
                    reverse_ip=reverse_ip, reverse_port=reverse_port),
                    ros_ws_abspath=ros_ws_abspath)
        
        # Internal Vars
        self.controllers_list = []
        
        # It doesnt use namespace
        self.robot_name_space = ""

        # We launch the init function of the Parent Class robot_gazebo_env.RobotGazeboEnv
        super(UR5LabSimEnv, self).__init__(controllers_list=self.controllers_list,
                                            robot_name_space=self.robot_name_space,
                                            reset_controls=False,
                                            start_init_physics_parameters=False,
                                            reset_world_or_sim="WORLD")

        self._gripper = rospy.Publisher('/gripper_controller/command', String, 
                                            queue_size=1)
                                            
        rospy.logdebug("UR5LabEnv unpausing")
        self.gazebo.unpauseSim()
                
        self._check_all_systems_ready()
        
        rospy.Subscriber("/joint_states", JointState, self._joints_state_callback)
        rospy.Subscriber("/move_group/status", GoalStatusArray, self._plan_status_feedback)
        self.last_contact_r = 0
        self.last_contact_l = 0
        
        self.uuid = roslaunch.rlutil.get_or_generate_uuid(None, False)
        roslaunch.configure_logging(self.uuid)
        ros_pack = rospkg.RosPack()
        ros_pack = ros_pack.get_path('ur5_moveit_config')
        ros_pack += '/launch/move_group.launch'
        self.launch = roslaunch.parent.ROSLaunchParent(self.uuid, [ros_pack])
        self.launch.start()
        rospy.loginfo("move_group started")
        
        self._move = MoveIt()
        
        rospy.logdebug("UR5LabEnv pausing")
        rospy.logdebug("Finished UR5LabEnv INIT...")
        
    # Methods needed by the RobotGazeboEnv
    # ----------------------------
    def close(self):
        self.launch.launch.shutdown()
        super().close()
        
    def _check_all_systems_ready(self):
        """
        Checks that all the sensors, publishers and other simulation systems are
        operational.
        """
        rospy.logdebug("UR5LabEnv check_all_systems_ready...")
        self._check_all_publishers_ready()
        self._check_all_sensors_ready()
        rospy.logdebug("END UR5LabEnv _check_all_systems_ready...")
        return True

    def _check_all_publishers_ready(self):
        rospy.logdebug("START ALL PUBLISHER")
        self._check_gripper_pub_ready()
        rospy.logdebug("All Publishers READY")
        
    def _check_gripper_pub_ready(self):
        rate = rospy.Rate(10)  # 10hz
        while (self._gripper.get_num_connections() == 0 and not rospy.is_shutdown()):
            rospy.logdebug(
                "No susbribers to _gripper yet so we wait and try again")
            try:
                rate.sleep()
            except rospy.ROSInterruptException:
                # This is to avoid error when world is rested, time when backwards.
                pass
        rospy.logdebug("_gripper Publisher Connected")
        
    def _check_all_sensors_ready(self):
        rospy.logdebug("START ALL SENSORS READY")
        self._check_joint_states_ready()
        rospy.logdebug("ALL SENSORS READY")

    def _JointState_numpy(self, data):
        data.name = list(data.name)
        data.position = list(data.position)
        finger_index = data.name.index('finger_joint')
        manipulator_index = [data.name.index('shoulder_pan_joint'), data.name.index('shoulder_lift_joint'),
                                data.name.index('elbow_joint'), data.name.index('wrist_1_joint'),
                                data.name.index('wrist_2_joint'), data.name.index('wrist_3_joint')]
                                
        manipulator_joints_position = [data.position[i] for i in manipulator_index]
                
        gripper_joints_position = data.position[finger_index]
        return manipulator_joints_position, gripper_joints_position
        
    def _check_joint_states_ready(self): #check published joint names and rearrange them for the move joint_goal
        self.joint_states = None
        self.gripper_states = None
        rospy.logdebug("Waiting for /joint_states to be READY...")
        while self.joint_states is None and not rospy.is_shutdown():
            try:
                self.joint_states, self.gripper_states = self._JointState_numpy(rospy.wait_for_message("/joint_states", JointState, timeout=2.0))
                rospy.logdebug("Current /joint_states READY=>")

            except:
                rospy.logerr("Current /joint_states not ready yet, retrying for getting joint_states")
        return self.joint_states

    def _check_no_motion_plan_ready(self):
        self.no_motion_plan = None
        rospy.logdebug("Waiting for Goal Status to be READY...")
        while self.no_motion_plan is None and not rospy.is_shutdown():
            try:
                self.no_motion_plan = self._no_motion_plan(rospy.wait_for_message("/move_group/status", GoalStatusArray, timeout=2.0))
                rospy.logdebug("Current Goal Status READY=>")

            except:
                rospy.logerr("Current Goal Status not ready yet, retrying for getting joint_states")
        return self.no_motion_plan

    def _no_motion_plan(self, data):
        if data.status_list:
            if data.status_list[-1].text in "No motion plan found. No execution attempted.":
                return True
        return False
        
    def _joints_state_callback(self, data):
        self.joint_states, self.gripper_states = self._JointState_numpy(data)
        return 

    def _plan_status_feedback(self, data):
        self.no_motion_plan = self._no_motion_plan(data)
        return

        
    # Methods that the TrainingEnvironment will need to define here as virtual
    # because they will be used in RobotGazeboEnv GrandParentClass and defined in the
    # TrainingEnvironment.
    # ----------------------------
    def _set_init_pose(self):
        """
        Sets the Robot in its init pose
        """
        raise NotImplementedError()

    def _init_env_variables(self):
        """
        Inits variables needed to be initialised each time 
        we reset at the start of an episode.
        """
        raise NotImplementedError()

    def _compute_reward(self, observations, done):
        """
        Calculates the reward to give based on the observations given.
        """
        raise NotImplementedError()

    def _set_action(self, action):
        """
        Applies the given action to the simulation.
        """
        raise NotImplementedError()

    def _get_obs(self):
        raise NotImplementedError()

    def _is_done(self, observations):
        """
        Checks if episode done based on observations given.
        """
        raise NotImplementedError()

    # Methods that the TrainingEnvironment will need.
    # ----------------------------
    def open_hand(self):
        """
        When called it opens robots hand
        """
        self._gripper.publish("open")
       
    def close_hand(self):
        """
        When called it closes robots hand
        """
        self._gripper.publish("close")
        
    def get_joint_states(self): 
        self.gazebo.unpauseSim()
        #joint = self._move.robot_joints()
        joint = self.joint_states
        return joint 
        
    def get_gripper_states(self):
        self.gazebo.unpauseSim()
        #joint = self._move.gripper_joints()
        joint = self.gripper_states
        return joint 
    
    def get_pc(self):
        return np.asarray(self.pc.points)
    
    def get_ee_pose(self): 
        """
        Returns the endeffector pose
        We make sure the simulation is unpaused because this calss is a service call.
        """
        rospy.logdebug("START getting endeffector pose")
        self.gazebo.unpauseSim()
        gripper_pose = self._move.ee_pose()
        rospy.logdebug("END getting endeffector pose")
        return gripper_pose   
        
    def get_ee_rpy(self): 
        rospy.logdebug("START getting endeffector rpy")
        self.gazebo.unpauseSim()
        gripper_rpy = self._move.ee_rpy()
        rospy.logdebug("END getting endeffector rpy")
        return gripper_rpy
        
    def move_ee(self, desired_pose): 
        """
        Moves the endeffector to the pose given,
        relative pose to world frame
        Currently it is only accepting position
        """
        pose_goal = Pose()
        pose_goal.position.x = desired_pose[0]
        pose_goal.position.y = desired_pose[1]
        pose_goal.position.z = desired_pose[2]
        pose_goal.orientation.x = desired_pose[3]
        pose_goal.orientation.y = desired_pose[4]
        pose_goal.orientation.z = desired_pose[5]
        pose_goal.orientation.w = desired_pose[6]
        
        
        result = self._move.ee_traj(pose_goal)
        return result
           
    def move_joints(self, joint_goal, wait=True): 
        """
        Moves the joints of the robot directly
        """
        result = self._move.joint_traj(joint_goal, wait)
        return result
    
    def check_pose(self, pose):
        pos = Pose()
        pos.position.x = pose[0]
        pos.position.y = pose[1]
        pos.position.z = pose[2]
        pos.orientation.x = pose[3]
        pos.orientation.y = pose[4]
        pos.orientation.z = pose[5]
        pos.orientation.w = pose[6]
        
        return self._move.check_pos(pos)[0]
    
    def get_observation(self):
        obs = self._get_obs()
        return obs

class MoveIt(object):
    def __init__(self):
        rospy.logdebug("===== MoveIt")
        moveit_commander.roscpp_initialize(sys.argv)
        self.robot = moveit_commander.RobotCommander()
        self.scene = moveit_commander.PlanningSceneInterface()
        self.move_group = moveit_commander.MoveGroupCommander("manipulator")
        self.move_group.set_num_planning_attempts(1)
        self.move_group.allow_replanning(False)
        self.eef_link = self.move_group.get_end_effector_link()
        self.joint_index()
        
        rospy.logdebug("============")

    def ee_traj(self, goal):
        self.move_group.clear_pose_targets()
        self.move_group.set_pose_target(goal)
        plan = self.move_group.go(wait=True)
        self.move_group.stop()
        self.move_group.clear_pose_targets()
        return plan

    def joint_traj(self, positions_array, wait=True):
        self.move_group.clear_pose_targets()   
        positions_array = [positions_array[i] for i in self.joint_name_index]
        result = self.move_group.go(positions_array, wait=wait)
        self.move_group.stop()
        return result
        
    def robot_joints(self):
        return self.move_group.get_current_joint_values()
        
    def check_pos(self, pose):
        return self.move_group.plan(pose)
    
    def joint_index(self):
        active_joints = self.move_group.get_active_joints()
        self.joint_name_index = [active_joints.index('shoulder_pan_joint'), active_joints.index('shoulder_lift_joint'),
                                active_joints.index('elbow_joint'), active_joints.index('wrist_1_joint'),
                                active_joints.index('wrist_2_joint'), active_joints.index('wrist_3_joint')]
        return
        
    def ee_pose(self):
        gripper_pose = self.move_group.get_current_pose().pose
        return gripper_pose

    def ee_rpy(self, request):
        gripper_rpy = self.move_group.get_current_rpy()
        return gripper_rpy
            



## -------------------------------------------------------------------------------------------------
## -- Project : MLPro - A Synoptic Framework for Standardized Machine Learning Tasks
## -- Package : mlpro.wrappers
## -- Module  : pettingzoo.py
## -------------------------------------------------------------------------------------------------
## -- History :
## -- yyyy-mm-dd  Ver.      Auth.    Description
## -- 2021-08-27  0.0.0     SY       Creation
## -- 2021-09-23  1.0.0     SY       Release of first version
## -- 2021-09-28  1.0.1     SY       WrEnvPZoo: implementation of method get_cycle_limits()
## -- 2021-09-29  1.1.0     SY       Change name:WrEnvPZoo to WrEnvPZOO2MLPro
## -- 2021-10-02  1.2.0     SY       New classes: WrEnvMLPro2PZoo, update _recognize_space() in WrEnvGYM2MLPro
## -- 2021-10-05  1.2.1     SY       Update following new attributes done and broken in State
## -- 2021-10-06  1.2.2     DA       Minor fixes
## -- 2021-10-07  1.2.3     SY       Update WrEnvMLPro2PZoo() 
## -- 2021-11-03  1.2.4     SY       Remove reset() on WrEnvPZOO2MLPro and WrEnvMLPro2PZoo to avoid double reset
## -- 2021-11-13  1.2.5     DA       Minor adjustments
## -- 2021-11-16  1.2.6     DA       Refactoring
## -- 2021-11-16  1.2.7     SY       Refactoring
## -- 2021-12-09  1.2.8     SY       Update process action procedure in WrEnvMLPro2PZoo()
## -- 2021-12-11  1.2.9     SY       Update WrEnvPZOO2MLPro() in setting up done flag
## -- 2021-12-21  1.3.0     DA       - Replaced 'done' by 'success' on mlpro functionality
## --                                - Optimized 'done' detection in both classes
## -- 2021-12-23  1.3.1     MRD      Remove adding self._num_cycle on simulate_reaction() due to 
## --                                EnvBase.process_actions() is already adding self._num_cycle
## -- 2022-01-20  1.3.2     SY       - Update PettingZoo2MLPro's reward type to C_TYPE_EVERY_AGENT 
## --                                - Update Wrapper MLPro2PettingZoo - Method step()
## -- 2022-01-21  1.3.3     SY       Class WrEnvPZOO2MLPro: 
## --                                - replace variable _reward to _last_reward 
## --                                Class WrEnvMLPro2PZoo:  
## --                                - refactored done detection 
## --                                - removed artifacts of cycle counting
## -- 2022-02-27  1.3.4     SY       Refactoring due to auto generated ID in class Dimension
## -- 2022-03-21  1.3.5     SY       Refactoring due to PettingZoo version 1.17.0
## -- 2022-05-20  1.3.6     SY       Refactoring: Action space boundaries in WrEnvPZOO2MLPro
## -- 2022-05-30  1.3.7     SY       Replace function env.seed(seed) to env.reset(seed=seed)
## -------------------------------------------------------------------------------------------------

"""
Ver. 1.3.7 (2022-05-30)
This module provides wrapper classes for reinforcement learning tasks.
"""


import gym
import numpy as np
from mlpro.rl.models import *
from mlpro.wrappers.openai_gym import WrEnvMLPro2GYM
from pettingzoo import AECEnv
from pettingzoo.utils import agent_selector
from pettingzoo.utils import wrappers





## -------------------------------------------------------------------------------------------------
## -------------------------------------------------------------------------------------------------
class WrEnvPZOO2MLPro(Environment):
    """
    This class is a ready to use wrapper class for Petting Zoo environments. 
    Objects of this type can be treated as an environment object. Encapsulated 
    petting zoo environment must be compatible to class pettingzoo.env.
    """

    C_TYPE        = 'Petting Zoo Env'

## -------------------------------------------------------------------------------------------------
    def __init__(self, p_zoo_env, p_state_space:MSpace=None, p_action_space:MSpace=None, p_logging=Log.C_LOG_ALL):
        """
        Parameters:
            p_pzoo_env      Petting Zoo environment object
            p_state_space   Optional external state space object that meets the
                            state space of the gym environment
            p_action_space  Optional external action space object that meets the
                            state space of the gym environment
            p_logging       Switch for logging
        """

        self._zoo_env     = p_zoo_env
        self.C_NAME       = 'Env "' + self._zoo_env.metadata['name'] + '"'

        Environment.__init__(self, p_mode=Environment.C_MODE_SIM, p_logging=p_logging)
        
        if p_state_space is not None: 
            self._state_space = p_state_space
        else:
            self._state_space = self._recognize_space(self._zoo_env.observation_spaces, "observation")
        
        if p_action_space is not None: 
            self._action_space = p_action_space
        else:
            self._action_space = self._recognize_space(self._zoo_env.action_spaces, "action")


## -------------------------------------------------------------------------------------------------
    def __del__(self):
        try:
            self._zoo_env.close()
        except:
            pass
        
        self.log(self.C_LOG_TYPE_I, 'Closed')


## -------------------------------------------------------------------------------------------------
    def _recognize_space(self, p_zoo_space, dict_name) -> ESpace:
        space = ESpace()
        
        if dict_name == "observation":
            space.add_dim(Dimension(p_name_short='0', p_base_set='DO'))
        elif dict_name == "action":
            for k in p_zoo_space:
                if isinstance(p_zoo_space[k], gym.spaces.Discrete):
                    space.add_dim(Dimension(p_name_short=k, p_base_set=Dimension.C_BASE_SET_Z,
                                            p_boundaries=[0, p_zoo_space[k].n]))
                elif isinstance(p_zoo_space[k], gym.spaces.Box):
                    shape_dim = len(p_zoo_space[k].shape)
                    for i in range(shape_dim):
                        for d in range(p_zoo_space[k].shape[i]):
                            space.add_dim(Dimension(p_name_short=str(d), p_base_set=Dimension.C_BASE_SET_R,
                                                    p_boundaries=[p_zoo_space[k].low[d], p_zoo_space[k].high[d]]))
                else:
                    space.add_dim(Dimension(p_name_short=k, p_base_set='DO'))
                
        return space


## -------------------------------------------------------------------------------------------------
    @staticmethod
    def setup_spaces():
        return None, None


## -------------------------------------------------------------------------------------------------
    def _reset(self, p_seed=None):

        # 1 Reset Zoo environment and determine initial state
        try:
            self._zoo_env.reset(seed=p_seed)
        except:
            self._zoo_env.seed(p_seed)
            self._zoo_env.reset()
        observation, _, _, _ = self._zoo_env.last()
        obs     = DataObject(observation)
        
        # 2 Create state object from Zoo observation
        state   = State(self._state_space)
        if isinstance(observation, dict):
            state.set_values(obs.get_data()['observation'])
        else:
            state.set_values(obs.get_data())
        self._set_state(state)


## -------------------------------------------------------------------------------------------------
    def simulate_reaction(self, p_state:State, p_action:Action) -> State:

        new_state = State(self._state_space)

        # 1 Convert action to Zoo syntax
        action_sorted = p_action.get_sorted_values()
        agent_num = 0
        
        for k in self._zoo_env.action_spaces:
            dtype = self._zoo_env.action_spaces[k].dtype
    
            if ( dtype == np.int32 ) or ( dtype == np.int64 ):
                action_sorted_agent = action_sorted[agent_num].round(0)
            else:
                action_sorted_agent = action_sorted[agent_num]
    
            action_zoo = action_sorted_agent.astype(self._zoo_env.action_spaces[k].dtype)
            
        # 2 Process step of Zoo environment that automatically switches control to the next agent.
            observation, reward_zoo, done, info = self._zoo_env.last()

            obs     = DataObject(observation)
            
            if done:
                self._zoo_env.step(None)
                new_state.set_terminal(True)

            else:
                try:
                    self._zoo_env.step(action_zoo)
                except:
                    self._zoo_env.step(np.atleast_1d(action_zoo))
            agent_num += 1

        # 3 Create state object from Zoo observation
            if isinstance(observation, dict):
                new_state.set_values(obs.get_data()['observation'])
            else:
                new_state.set_values(obs.get_data())

        # 4 Create and store reward object
        self._last_reward = Reward(Reward.C_TYPE_EVERY_AGENT)
        for key in self._zoo_env.rewards.keys():
            self._last_reward.add_agent_reward(key, self._zoo_env.rewards.get(key))

        return new_state


## -------------------------------------------------------------------------------------------------
    def compute_reward(self, p_state_old:State=None, p_state_new:State=None) -> Reward:
        if ( p_state_old is not None ) or ( p_state_new is not None ):
            raise NotImplementedError

        return self._last_reward


## -------------------------------------------------------------------------------------------------
    def compute_success(self, p_state:State) -> bool:
        return self.get_success()


## -------------------------------------------------------------------------------------------------
    def compute_broken(self, p_state:State) -> bool:
        return self.get_broken()


## -------------------------------------------------------------------------------------------------
    def init_plot(self, p_figure=None):
        self._zoo_env.render()


## -------------------------------------------------------------------------------------------------
    def update_plot(self):
        self._zoo_env.render()


## -------------------------------------------------------------------------------------------------
    def get_cycle_limit(self):
        try:
            return self._zoo_env.env.env.max_cycles
        except:
            return self.C_CYCLE_LIMIT





## -------------------------------------------------------------------------------------------------
## -------------------------------------------------------------------------------------------------
class WrEnvMLPro2PZoo():
    """
    This class is a ready to use wrapper class for MLPro to PettingZoo environments. 
    Objects of this type can be treated as an AECEnv object. Encapsulated 
    MLPro environment must be compatible to class Environment.
    To be noted, this wrapper is not capable for parallel environment yet.
    """

    C_TYPE        = 'MLPro to PZoo Env'

## -------------------------------------------------------------------------------------------------
    def __init__(self, p_mlpro_env, p_num_agents, p_state_space:MSpace=None, p_action_space:MSpace=None):
        """
        Parameters:
            p_mlpro_env     MLPro's Environment object
            p_num_agents    Number of Agents
            p_state_space   Optional external state space object that meets the
                            state space of the MLPro environment
            p_action_space  Optional external action space object that meets the
                            state space of the MLPro environment
        """
        
        self.pzoo_env   = self.raw_env(p_mlpro_env, p_num_agents, p_state_space, p_action_space)
        self.pzoo_env   = wrappers.CaptureStdoutWrapper(self.pzoo_env)
        self.pzoo_env   = wrappers.OrderEnforcingWrapper(self.pzoo_env)


## -------------------------------------------------------------------------------------------------
    class raw_env(AECEnv):
        metadata = {'render_modes': ['human', 'ansi'], "name": "pzoo_custom"}

## -------------------------------------------------------------------------------------------------
        def __init__(self, p_mlpro_env, p_num_agents, p_state_space:MSpace=None, p_action_space:MSpace=None):
            self._mlpro_env             = p_mlpro_env
            self.possible_agents        = [str(r) for r in range(p_num_agents)]
            self.agent_name_mapping     = dict(zip(self.possible_agents, list(range(len(self.possible_agents)))))
            
            if p_state_space is not None: 
                self.observation_spaces = p_state_space
            else:
                self.observation_spaces = self._recognize_space(self._mlpro_env.get_state_space())
            
            if p_action_space is not None: 
                self.action_spaces      = p_action_space
            else:
                self.action_spaces      = self._recognize_space(self._mlpro_env.get_action_space())
            
            self.first_refresh          = True
        

## -------------------------------------------------------------------------------------------------
        def _recognize_space(self, p_mlpro_space):
            space           = WrEnvMLPro2GYM.recognize_space(p_mlpro_space)
            setup_space     = {agent: space for agent in self.possible_agents}
                
            return setup_space


## -------------------------------------------------------------------------------------------------
        def step(self, action):
            if self.dones[self.agent_selection]:
                return self._was_done_step(action)
            
            agent = self.agent_selection
            self._cumulative_rewards[agent] = 0
            cycle_limit = self._mlpro_env.get_cycle_limit()
            
            self.state[self.agent_selection] = action[int(self.agent_selection)]
            
            if agent == self.possible_agents[-1]:
                _action     = Action()
                idx         = self._mlpro_env.get_action_space().get_num_dim()
                if isinstance(self.observation_spaces, gym.spaces.Discrete):
                    action = np.array([action])
                for i in range(idx):
                    _act_set    = Set()
                    _act_set.add_dim(Dimension('action_'+str(i)))
                    _act_elem   = Element(_act_set)
                    _ids = _act_elem.get_dim_ids()
                    _act_elem.set_value(_ids[0], self.state[self.possible_agents[i]])
                    _action.add_elem(self.possible_agents[i], _act_elem)
                
                self._mlpro_env.process_action(_action)
                for i in range(idx):
                    self.rewards[self.possible_agents[i]] = self._mlpro_env.compute_reward().get_agent_reward(self.possible_agents[i]).item()
                    if not self.rewards[self.possible_agents[i]]:
                        self.rewards[self.possible_agents[i]] = 0
            
                if self._mlpro_env.get_state().get_terminal():
                    self.dones = {agent: True for agent in self.agents}
                    
                for i in self.agents:
                    self.observations[i] = self.state[self.agents[1-int(i)]]
            
            else:
                self._clear_rewards()
                
            self.agent_selection = self._agent_selector.next()
            
            self._accumulate_rewards()


## -------------------------------------------------------------------------------------------------
        def observe(self, agent_id):
            # highly recommended to reimplement this function, since the states for
            # each agent in different environments can not be standardized
            dim     = len(self._mlpro_env.get_state().get_dim_ids())
            state   = []
            for i in range(dim):
                state.append(self._mlpro_env.get_state().get_values()[i].item())
            return np.array(state, dtype=np.float32)


## -------------------------------------------------------------------------------------------------
        def reset(self):
            self.agents = self.possible_agents[:]
            self.rewards = {agent: 0 for agent in self.agents}
            self._cumulative_rewards = {agent: 0 for agent in self.agents}
            self.dones = {agent: False for agent in self.agents}
            self.infos = {agent: {} for agent in self.agents}
            self.state = {agent: None for agent in self.agents}
            self.observations = {agent: None for agent in self.agents}
            
            self._mlpro_env.reset()
            
            self._agent_selector = agent_selector(self.agents)
            self.agent_selection = self._agent_selector.next()


## -------------------------------------------------------------------------------------------------
        def render(self, mode='human'):
            try:
                if self.first_refresh:
                    self._mlpro_env.init_plot()
                    self.first_refresh = False
                else:
                    self._mlpro_env.update_plot()
                return True
            except:
                return False


## -------------------------------------------------------------------------------------------------
        def close(self):
            self._mlpro_env.__del__()


## -------------------------------------------------------------------------------------------------
## -- Project : FH-SWF Automation Technology - Common Code Base (CCB)
## -- Package : mlpro
## -- Module  : test_policy_wrapper
## -------------------------------------------------------------------------------------------------
## -- History :
## -- yyyy-mm-dd  Ver.      Auth.    Description
## -- 2021-09-11  1.0.0     MRD      Creation
## -- 2021-09-21  1.0.0     MRD      Release First Version
## -- 2021-10-27  1.0.1     MRD      Added Policy Loss Check between Native and Wrapper
## -- 2021-12-08  1.0.2     DA       Refactoring
## -- 2021-12-20  1.0.3     DA       Refactoring
## -- 2022-01-18  2.0.0     MRD      Add Off Policy Algorithm into the test
## -- 2022-01-21  2.0.1     MRD      Include RobotHTM as the continues action envrionment
## -------------------------------------------------------------------------------------------------

"""
Ver. 2.0.1 (2022-01-21)

Unit test classes for environment.
"""

from numpy import empty
import pytest
import gym
import torch
from mlpro.rl.models import *
from mlpro.wrappers.openai_gym import WrEnvGYM2MLPro, WrEnvMLPro2GYM
from mlpro.rl.pool.envs.robotinhtm import RobotHTM
from mlpro.wrappers.sb3 import WrPolicySB32MLPro
from stable_baselines3 import A2C, PPO, DQN, DDPG, SAC
from stable_baselines3.common.on_policy_algorithm import OnPolicyAlgorithm
from stable_baselines3.common.callbacks import BaseCallback


## -------------------------------------------------------------------------------------------------
@pytest.mark.parametrize("env_cls", [PPO, A2C, DQN, DDPG, SAC])
def test_sb3_policy_wrapper(env_cls):
    buffer_size = 100
    policy_kwargs_on = dict(activation_fn=torch.nn.Tanh,
                            net_arch=[dict(pi=[128, 128], vf=[128, 128])])

    policy_kwargs_off = dict(activation_fn=torch.nn.ReLU,
                             net_arch=[10])

    class MyScenario(RLScenario):

        C_NAME = 'Matrix'

        def _setup(self, p_mode, p_ada, p_logging):
            class CustomWrapperFixedSeed(WrEnvGYM2MLPro):
                def _reset(self, p_seed=None):
                    self.log(self.C_LOG_TYPE_I, 'Reset')
                    self._num_cycles = 0

                    # 1 Reset Gym environment and determine initial state
                    observation = self._gym_env.reset()
                    obs = DataObject(observation)

                    # 2 Create state object from Gym observation
                    state = State(self._state_space)
                    state.set_values(obs.get_data())
                    self._set_state(state)

            if issubclass(env_cls, OnPolicyAlgorithm):
                # 1 Setup environment
                self._env = RobotHTM(p_reset_seed=False, p_target_mode="fix", p_logging=False)
                policy_sb3 = env_cls(
                    policy="MlpPolicy",
                    env=None,
                    n_steps=buffer_size,
                    _init_setup_model=False,
                    policy_kwargs=policy_kwargs_on,
                    verbose=0,
                    seed=2)
            else:
                if issubclass(env_cls, DQN):
                    # 1 Setup environment
                    gym_env = gym.make('CartPole-v1')
                    gym_env.seed(2)
                    self._env = CustomWrapperFixedSeed(gym_env, p_logging=False)
                else:
                    self._env = RobotHTM(p_reset_seed=False, p_target_mode="fix", p_logging=False)

                policy_sb3 = env_cls(
                    policy="MlpPolicy",
                    env=None,
                    buffer_size=1000000,
                    _init_setup_model=False,
                    policy_kwargs=policy_kwargs_off,
                    gradient_steps=1,
                    train_freq=4,
                    learning_starts=0,
                    verbose=0,
                    seed=2)

            class TestWrPolicySB32MLPro(WrPolicySB32MLPro):
                """
                Custom Class for logging the loss
                """

                def __init__(self, p_sb3_policy, p_cycle_limit, p_observation_space, p_action_space, p_ada=True,
                             p_logging=True):
                    super().__init__(p_sb3_policy, p_cycle_limit, p_observation_space, p_action_space, p_ada=p_ada,
                                     p_logging=p_logging)
                    self.loss_cnt = []

                def _adapt_off_policy(self, *p_args) -> bool:
                    if super()._adapt_off_policy(*p_args):
                        if isinstance(self.sb3, DQN):
                            self.loss_cnt.append(self.sb3.logger.name_to_value["train/loss"])
                        elif isinstance(self.sb3, DDPG):
                            self.loss_cnt.append(self.sb3.logger.name_to_value["train/critic_loss"])
                        elif isinstance(self.sb3, SAC):
                            self.loss_cnt.append(self.sb3.logger.name_to_value["train/critic_loss"])
                        return True
                    return False

                def _adapt_on_policy(self, *p_args) -> bool:
                    if super()._adapt_on_policy(*p_args):
                        # Log the Loss
                        if isinstance(self.sb3, PPO):
                            self.loss_cnt.append(self.sb3.logger.name_to_value["train/policy_gradient_loss"])
                        elif isinstance(self.sb3, A2C):
                            self.loss_cnt.append(self.sb3.logger.name_to_value["train/policy_loss"])
                        return True
                    return False

            # 3 Wrap the policy
            self.policy_wrapped = TestWrPolicySB32MLPro(
                p_sb3_policy=policy_sb3,
                p_cycle_limit=self._cycle_limit,
                p_observation_space=self._env.get_state_space(),
                p_action_space=self._env.get_action_space(),
                p_ada=p_ada,
                p_logging=p_logging)

            # 4 Setup standard single-agent with own policy
            return Agent(
                p_policy=self.policy_wrapped,
                p_envmodel=None,
                p_name='Smith',
                p_ada=p_ada,
                p_logging=p_logging
            )

    # 3 Instantiate training
    training = RLTraining(
        p_scenario_cls=MyScenario,
        p_cycle_limit=1200,
        p_stagnation_limit=0,
        p_collect_states=True,
        p_collect_actions=True,
        p_collect_rewards=True,
        p_collect_training=True,
        p_logging=False
    )

    # 4 Train
    training.run()

    class CustomCallback(BaseCallback):

        def __init__(self, p_verbose=0):
            super(CustomCallback, self).__init__(p_verbose)
            self.update = 0
            self.loss_cnt = []

        def _on_rollout_start(self) -> None:
            if self.update != 0:
                if isinstance(self.locals.get("self"), PPO):
                    self.loss_cnt.append(self.locals.get("self").logger.name_to_value["train/policy_gradient_loss"])
                elif isinstance(self.locals.get("self"), A2C):
                    self.loss_cnt.append(self.locals.get("self").logger.name_to_value["train/policy_loss"])
                elif isinstance(self.locals.get("self"), DQN):
                    self.loss_cnt.append(self.locals.get("self").logger.name_to_value["train/loss"])
                elif isinstance(self.locals.get("self"), DDPG):
                    self.loss_cnt.append(self.locals.get("self").logger.name_to_value["train/critic_loss"])
                elif isinstance(self.locals.get("self"), SAC):
                    self.loss_cnt.append(self.locals.get("self").logger.name_to_value["train/critic_loss"])

        def _on_step(self) -> bool:
            return super()._on_step()

        def _on_rollout_end(self) -> None:
            self.update += 1

    if issubclass(env_cls, OnPolicyAlgorithm):
        # 1 Setup environment
        env = RobotHTM(p_reset_seed=False, p_target_mode="fix", p_logging=False)
        gym_env = WrEnvMLPro2GYM(env)
        policy_sb3 = env_cls(
            policy="MlpPolicy",
            env=gym_env,
            n_steps=buffer_size,
            verbose=0,
            policy_kwargs=policy_kwargs_on,
            seed=2)
    else:
        if issubclass(env_cls, DQN):
            # 1 Setup environment
            gym_env = gym.make('CartPole-v1')
            gym_env.seed(2)
        else:
            env = RobotHTM(p_reset_seed=False, p_target_mode="fix", p_logging=False)
            gym_env = WrEnvMLPro2GYM(env)
        policy_sb3 = env_cls(
            policy="MlpPolicy",
            env=gym_env,
            buffer_size=1000000,
            verbose=0,
            gradient_steps=1,
            train_freq=4,
            learning_starts=0,
            policy_kwargs=policy_kwargs_off,
            seed=2)

    cus_callback = CustomCallback()
    policy_sb3.learn(total_timesteps=1200, callback=cus_callback)

    assert cus_callback.loss_cnt is not empty, "No Loss on Native"
    assert training.get_scenario().policy_wrapped.loss_cnt is not empty, "No Loss on Wrapper"
    length = min(len(cus_callback.loss_cnt), len(training.get_scenario().policy_wrapped.loss_cnt))
    assert np.linalg.norm(np.array(cus_callback.loss_cnt[:length]) - np.array(
        training.get_scenario().policy_wrapped.loss_cnt[:length])) == 0.0, "Mismatch Native and Wrapper"


from gym import spaces
import my_robot_env
from gym.envs.registration import register
import rospy

# The path is __init__.py of openai_ros, where we import the MovingCubeOneDiskWalkEnv directly
timestep_limit_per_episode = 1000 # Can be any Value

register(
        id='MyTrainingEnv-v0',
        entry_point='template_my_training_env:MovingCubeOneDiskWalkEnv',
        timestep_limit=timestep_limit_per_episode,
    )

class MyTrainingEnv(cube_single_disk_env.MyRobotEnv):
    def __init__(self):
        
        # Only variable needed to be set here
        number_actions = rospy.get_param('/my_robot_namespace/n_actions')
        self.action_space = spaces.Discrete(number_actions)
        
        # This is the most common case of Box observation type
        high = numpy.array([
            obs1_max_value,
            obs12_max_value,
            ...
            obsN_max_value
            ])
            
        self.observation_space = spaces.Box(-high, high)
        
        # Variables that we retrieve through the param server, loded when launch training launch.
        


        # Here we will add any init functions prior to starting the MyRobotEnv
        super(MyTrainingEnv, self).__init__()


    def _set_init_pose(self):
        """Sets the Robot in its init pose
        """
        # TODO

    def _init_env_variables(self):
        """
        Inits variables needed to be initialised each time we reset at the start
        of an episode.
        :return:
        """
        # TODO


    def _set_action(self, action):
        """
        Move the robot based on the action variable given
        """
        # TODO: Move robot

    def _get_obs(self):
        """
        Here we define what sensor data of our robots observations
        To know which Variables we have acces to, we need to read the
        MyRobotEnv API DOCS
        :return: observations
        """
        # TODO
        return observations

    def _is_done(self, observations):
        """
        Decide if episode is done based on the observations
        """
        # TODO
        return done

    def _compute_reward(self, observations, done):
        """
        Return the reward based on the observations given
        """
        # TODO
        return reward
        
    # Internal TaskEnv Methods




## -------------------------------------------------------------------------------------------------
## -- Project : MLPro - A Synoptic Framework for Standardized Machine Learning Tasks
## -- Package : mlpro
## -- Module  : howto_bf_001_logging.py
## -------------------------------------------------------------------------------------------------
## -- History :
## -- yyyy-mm-dd  Ver.      Auth.    Description
## -- 2021-10-07  1.0.0     DA       Creation
## -- 2021-11-03  1.1.0     DA       Introduction of new log type C_LOG_TYPE_S for success messages
## -- 2021-11-13  1.1.1     DA       Refactoring
## -------------------------------------------------------------------------------------------------

"""
Ver. 1.1.1 (2021-11-13)

This module demonstrates the Log class functionality.
"""


from mlpro.bf.various import Log




# 1 Reuse logging property in your own class by inheriting from class Log
class MyClass(Log):

    # These constants are inherited fromm class Log and will be logged in every log line...
    C_TYPE  = 'Demo class'      
    C_NAME  = 'MyClass'

    def __init__(self, p_logging=True):
        # The constructor of class Log initializes the internal logging and writes the first line...
        super().__init__(p_logging=p_logging)


    def my_method(self):
        # The log types I/E/W are also inherited from class Log...
        self.log(self.C_LOG_TYPE_I, 'Let me tell you what\'s going on...')
        self.log(self.C_LOG_TYPE_W, 'Something is weird...')
        self.log(self.C_LOG_TYPE_E, 'And here something failed...')
        self.log(self.C_LOG_TYPE_I, 'But don\'t worry. Everything is fine. It\'s just a demo:)')
        self.log(self.C_LOG_TYPE_S, 'This method terminated successfully!\n')




if __name__ == "__main__":
    
    # 2 Log everything inside your class... 
    print('\n--\n-- Example 1: By default everything is logged...\n--\n')
    mc = MyClass(p_logging=Log.C_LOG_ALL)
    mc.my_method()


    # 3 Log nothing inside your class
    print('\n--\n-- Example 2: Now logging is switched off...\n--\n')
    mc.switch_logging(Log.C_LOG_NOTHING)
    mc.my_method()


    # 4 Log warnings and errors only
    print('\n--\n-- Example 3: Only warnings and errors are logged...\n--\n')
    mc.switch_logging(Log.C_LOG_WE)
    mc.my_method()

    # 5 Log errors only
    print('\n--\n-- Example 4: Only errors are logged...\n--\n')
    mc.switch_logging(Log.C_LOG_E)
    mc.my_method()

    # 6 Log everything again
    print('\n--\n-- Example 5: Everything is logged again...\n--\n')
    mc.switch_logging(Log.C_LOG_ALL)
    mc.my_method()



## -------------------------------------------------------------------------------------------------
## -- Project : MLPro - A Synoptic Framework for Standardized Machine Learning Tasks
## -- Package : mlpro
## -- Module  : howto_rl_014_advanced_training_with_stagnation_detection.py
## -------------------------------------------------------------------------------------------------
## -- History :
## -- yyyy-mm-dd  Ver.      Auth.    Description
## -- 2021-12-11  0.0.0     DA       Creation
## -- 2021-12-12  1.0.0     DA       Released first version
## -- 2022-02-04  1.1.0     DA       Introduction of parameter p_stagnation_entry
## -- 2022-02-10  1.2.0     DA       Introduction of parameter p_end_at_stagnation
## -- 2022-02-27  1.2.1     SY       Refactoring due to auto generated ID in class Dimension
## -- 2022-05-19  1.2.2     SY       Utilize RandomGenerator
## -------------------------------------------------------------------------------------------------

"""
Ver. 1.2.2 (2022-05-19)

This module demonstrates advanced training with evaluation and stagnation detection.
"""


from mlpro.rl.models import *
from mlpro.rl.pool.envs.multicartpole import MultiCartPole
import random
from pathlib import Path
from mlpro.rl.pool.policies.randomgenerator import RandomGenerator




# 1 Implement your own RL scenario
class MyScenario (RLScenario):

    C_NAME      = 'Matrix'

    def _setup(self, p_mode, p_ada, p_logging):

        # 1 Setup Multi-Agent Environment (consisting of 3 OpenAI Gym Cartpole envs)
        self._env   = MultiCartPole(p_num_envs=3, p_reward_type=Reward.C_TYPE_EVERY_AGENT, p_logging=p_logging)


        # 2 Setup Multi-Agent 

        # 2.1 Create empty Multi-Agent
        multi_agent = MultiAgent(
            p_name='Smith',
            p_ada=True,
            p_logging=p_logging
        )

        # 2.2 Add Single-Agent #1 with own policy (controlling sub-environment #1)
        ss_ids = self._env.get_state_space().get_dim_ids()
        as_ids = self._env.get_action_space().get_dim_ids()
        multi_agent.add_agent(
            p_agent=Agent(
                p_policy=RandomGenerator(
                    p_observation_space=self._env.get_state_space().spawn([ss_ids[0],ss_ids[1],ss_ids[2],ss_ids[3]]),
                    p_action_space=self._env.get_action_space().spawn([as_ids[0]]),
                    p_buffer_size=1,
                    p_ada=True,
                    p_logging=p_logging
                ),
                p_envmodel=None,
                p_name='Smith-1',
                p_id=0,
                p_ada=True,
                p_logging=p_logging
            ),
            p_weight=0.3
        )

        # 2.3 Add Single-Agent #2 with own policy (controlling sub-environments #2,#3)
        multi_agent.add_agent(
            p_agent=Agent(
                p_policy=RandomGenerator(
                    p_observation_space=self._env.get_state_space().spawn([ss_ids[4],ss_ids[5],ss_ids[6],ss_ids[7],ss_ids[8],ss_ids[9],ss_ids[10],ss_ids[11]]),
                    p_action_space=self._env.get_action_space().spawn([as_ids[1],as_ids[2]]),
                    p_buffer_size=1,
                    p_ada=True,
                    p_logging=p_logging
                ),
                p_envmodel=None,
                p_name='Smith-2',
                p_id=1,
                p_ada=True,
                p_logging=p_logging
            ),
            p_weight=0.7
        )

        # 2.4 Adaptive ML model (here: our multi-agent) is returned
        return multi_agent



# 3 Create scenario and start training

if __name__ == "__main__":
    # 3.1 Parameters for demo mode
    cycle_limit         = 1500
    adaptation_limit    = 50
    stagnation_limit    = 5
    stagnation_entry    = 3
    end_at_stagnation   = True
    eval_frequency      = 10
    eval_grp_size       = 5
    logging             = Log.C_LOG_WE
    visualize           = True
    path                = str(Path.home())
 
else:
    # 3.2 Parameters for internal unit test
    cycle_limit         = 50
    adaptation_limit    = 5
    stagnation_limit    = 5
    stagnation_entry    = 1
    end_at_stagnation   = True
    eval_frequency      = 2
    eval_grp_size       = 1
    logging             = Log.C_LOG_NOTHING
    visualize           = False
    path                = None


# 3.3 Create and run training object
training = RLTraining(
        p_scenario_cls=MyScenario,
        p_cycle_limit=cycle_limit,
        p_adaptation_limit=adaptation_limit,
        p_eval_frequency=eval_frequency,
        p_eval_grp_size=eval_grp_size,
        p_stagnation_limit=stagnation_limit,
        p_stagnation_entry=stagnation_entry,
        p_end_at_stagnation=end_at_stagnation,
        p_score_ma_horizon=3,
        p_success_ends_epi=True,
        p_path=path,
        p_visualize=visualize,
        p_logging=logging )

training.run()


## -------------------------------------------------------------------------------------------------
## -- Project : MLPro - A Synoptic Framework for Standardized Machine Learning Tasks
## -- Package : mlpro.rl.envs
## -- Module  : bglp.py
## -------------------------------------------------------------------------------------------------
## -- History :
## -- yyyy-mm-dd  Ver.      Auth.    Description
## -- 2021-06-07  0.0.0     SY       Creation
## -- 2021-08-31  1.0.0     SY       Release of first version
## -- 2021-09-01  1.0.1     SY       Minor improvements, code cleaning, add descriptions
## -- 2021-09-06  1.0.2     SY       Minor improvements, combine bglp and BGLP classes
## -- 2021-09-11  1.0.3     MRD      Change Header information to match our new library name
## -- 2021-10-02  1.0.4     SY       Minor change
## -- 2021-10-05  1.0.5     SY       Update following new attributes done and broken in State
## -- 2021-10-07  2.0.0     SY       Enable batch production scenario
## -- 2021-10-25  2.0.1     SY       Add scientific references related to the Environment
## -- 2021-11-15  2.1.0     DA       Refactoring
## -- 2021-11-16  2.1.1     SY       Update following model improvements
## -- 2021-11-16  2.1.2     SY       Add data storing for overflow, demand, energy
## -- 2021-11-17  2.1.3     SY       Random initial states
## -- 2021-11-21  2.1.4     SY       Remove dependency from torch
## -- 2021-11-26  2.1.5     SY       Update reward type
## -- 2021-12-03  2.1.6     DA       Refactoring
## -- 2021-12-09  2.1.7     SY       Clean code assurance
## -- 2021-12-19  2.1.8     DA       Replaced 'done' by 'success'
## -- 2021-12-21  2.1.9     DA       Class BGLP: renamed method reset() to _reset()
## -- 2022-01-21  2.2.0     SY       Add cycle_limit as an input parameter
## -- 2022-01-24  2.2.1     SY       Update seeding procedure, refactoring _reset()
## -- 2022-02-25  2.2.2     SY       Refactoring due to auto generated ID in class Dimension
## -- 2022-05-23  2.2.3     SY       Bug fixing: Reward computation
## -- 2022-05-30  2.2.4     SY       Replace 'energy' related parameters to 'power'
## -- 2022-06-14  2.2.5     SY       Add termination condition for batch production scenario
## -------------------------------------------------------------------------------------------------

"""
Ver. 2.2.5 (2022-06-14)

This module provides an RL environment of Bulk Good Laboratory Plant (BGLP).
"""


from mlpro.rl.models import *
from mlpro.bf.various import *
import numpy as np
import random




## -------------------------------------------------------------------------------------------------
## -------------------------------------------------------------------------------------------------
class Actuator:
    """
    This class serves as a parent class of different types of actuators, which provides the main 
    attributes of an actuator in the BGLP environment.

    Parameters
    ----------
    minpower : float
        minimum power of an actuator.
    maxpower : float
        maximum power of an actuator.
    minaction : float
        minimum action of an actuator.
    maxaction : float
        maximum action of an actuator.
    masscoeff : float
        mass transport coefficient of an actuator.
        
    Attributes
    ----------
    reg_a : list of objects
        list of existing actuators in the environment.
    idx_a : int
        length of reg_a.
    power_max : float
        maximum power of an actuator.
    power_min : float
        minimum power of an actuator.
    power_coeff : float
        power coefficient of an actuator, if necessary.
    action_max : float
        maximum action of an actuator.
    action_min : float
        minimum action of an actuator.
    mass_coeff : float
        mass transport coefficient of an actuator.
    t_activated : float
        a time indicator about an actuator is activated.
    t_end : float
        a time indicator about the end of an activation sequence of the actuator.
    status : bool
        status of an actuator, false means inactive and true means active.
    cur_mass_transport : float
        current transported mass of an actuator.
    cur_power : float
        current power consumption of an actuator.
    cur_action : float
        current taken action of an actuator in RL context.
    cur_speed : float
        current speed of an actuator.
    type_ : str
        a short name for an actuator, usually 3 capital letters (e.g VAC, BLT).
    """
    reg_a               = []
    idx_a               = 0
    power_max           = 0
    power_min           = 0
    power_coeff         = 0
    action_max          = 0
    action_min          = 0
    mass_coeff          = 0
    t_activated         = 0
    t_end               = 0
    status              = False
    cur_mass_transport  = 0
    cur_power           = 0
    cur_action          = 0
    cur_speed           = 0
    type_               = ""
    
## -------------------------------------------------------------------------------------------------
    def __init__(self, minpower, maxpower, minaction, maxaction, masscoeff):
        self.idx_a              = len(self.reg_a)
        self.reg_a.append(self)
        self.power_max          = maxpower
        self.power_min          = minpower
        self.action_min         = minaction
        self.action_max         = maxaction
        self.mass_coeff         = masscoeff
        self.cur_mass_transport = 0
        self.cur_power          = 0
        self.cur_action         = 0
        self.cur_speed          = 0




## -------------------------------------------------------------------------------------------------
## -------------------------------------------------------------------------------------------------
class VacuumPump (Actuator):
    """
    This class inherits Actuator class and serves as a child class of Actuator.
    This class represents a type of actuators in the BGLP environment, namely Vacuum Pump.
    Vacuum Pumps are mostly used to transport material from mini hoppers to silos.
    However, the parameter of each vacuum pump can be dissimilar to each other based on their settings.

    Parameters
    ----------
    name : str
        specific name or id of a vacuum pump (e.g. Vac_A, etc.).
    minpower : float
        minimum power of a vacuum pump.
    maxpower : float
        maximum power of a vacuum pump.
    minaction : float
        minimum action of a vacuum pump.
    maxaction : float
        maximum action of a vacuum pump.
    masscoeff : float
        mass transport coefficient of a vacuum pump.
        
    Attributes
    ----------
    reg_v : list of objects
        list of existing vacuum pumps.
    idx_v : int
        length of reg_v.
    name : str
        specific name or id of a vacuum pump.
    t_end_max : float
        maximum end of activation time of a vacuum pump with respect to current time.
    """
    reg_v       = []
    idx_v       = 0
    name        = ""
    t_end_max   = 0
    
## -------------------------------------------------------------------------------------------------
    def __init__(self, name, minpower, maxpower, minaction, maxaction, masscoeff):    
        Actuator.__init__(self, minpower, maxpower, minaction, maxaction, masscoeff)
        self.idx_v  = len(self.reg_v)
        self.reg_v.append(self)
        self.name   = name
        self.type_  = "VAC"
        

## -------------------------------------------------------------------------------------------------
    def start_t(self, now, duration, overwrite = False): 
        """
        This method calculates the activation time and the end of activation time of the vacuum pump.
        This method is called, if the vacuum pump would like to be activated or updated.

        Parameters
        ----------
        now : float
            current time of the system.
        duration : float
            duration of the vacuum pump being activated or the action by an agent in RL context.
        overwrite : bool, optional
            To indicate whether the current operation can be overwritten or not.
        """
        if self.status == False:
            self.t_activated    = now
            self.t_end          = now + duration * self.action_max
            self.t_end_max      = now + self.action_max
            self.cur_action     = duration * self.action_max
            self.status         = True
        else:
            if overwrite:
                self.cur_action = duration * self.action_max
                self.t_end      = now + duration * self.action_max    
                self.status     = True
                

## -------------------------------------------------------------------------------------------------
    def calc_mass(self, now):
        """
        This method calculates the transported mass flow by the vacuum pump for a time step.

        Parameters
        ----------
        now : float
            current time of the system.

        Returns
        -------
        cur_mass_transport : float
            current transported mass.

        """
        if self.status == True:
            t_diff = now - (self.t_activated + self.action_min)
            if t_diff >= 0:
                self.cur_mass_transport = (2*self.mass_coeff[1]) + self.mass_coeff[0]
        return self.cur_mass_transport        
        

## -------------------------------------------------------------------------------------------------
    def calc_power(self):
        """
        This method calculates the power consumption of a vacuum pump.

        Returns
        -------
        cur_power : float
            current power consumption.

        """
        if self.status == True:
            self.cur_power = self.power_max
        return self.cur_power / 1000.0
        

## -------------------------------------------------------------------------------------------------
    def update(self, now):
        """
        This method calculates whether a vacuum pump must be deactived or not.

        Parameters
        ----------
        now : float
            current time of the system.

        """
        if self.status == True:
            if self.t_end < now or self.t_end_max < now:
                self.deactivate()
    

## -------------------------------------------------------------------------------------------------
    def deactivate(self):
        """
        This method is used to deactivate a vacuum pump.
        """
        self.status             = False
        self.cur_mass_transport = 0
        self.cur_power          = 0




## -------------------------------------------------------------------------------------------------
## -------------------------------------------------------------------------------------------------
class Belt(Actuator):
    """
    This class inherits Actuator class and serves as a child class of Actuator.
    This class represents a type of actuators in the BGLP environment, namely Belt.
    This class can be used for Conveyor Belt, Rotary Feeder, Vibratory Conveyor, or similar type of actuators.
    Belts are mostly used to transport material from silos to hoppers.
    However, the parameter of each actuator can be dissimilar to each other based on their settings.

    Parameters
    ----------
    name : str
        specific name or id of an actuator (e.g. Belt_A, etc.).
    actiontype : str
        "C" for continuous action, "B" for binary action.
    minpower : float
        minimum power of an actuator.
    maxpower : float
        maximum power of an actuator.
    minaction : float
        minimum action of an actuator.
    maxaction : float
        maximum action of an actuator.
    masscoeff : float
        mass transport coefficient of an actuator.
        
    Attributes
    ----------
    reg_b : list of objects
        list of existing actuators.
    idx_b : int
        length of reg_b.
    name : str
        specific name or id of an actuator.
    actiontype : str
        "C" for continuous action, "B" for binary action.
    speed : float
        speed of an actuator.
    """
    reg_b       = []
    idx_b       = 0
    name        = ""
    actiontype  = ""
    speed       = 0
    
## -------------------------------------------------------------------------------------------------
    def __init__(self, name, actiontype, minpower, maxpower, minaction, maxaction, masscoeff):
        Actuator.__init__(self, minpower, maxpower, minaction, maxaction, masscoeff)
        self.idx_b = len(self.reg_b)
        self.reg_b.append(self)
        self.name = name
        self.actiontype = actiontype
        self.speed = 0
        self.type_ = "BLT"
        

## -------------------------------------------------------------------------------------------------
    def start_t(self, now, duration, speed, overwrite = False):
        """
        This method calculates the activation time and the end of activation time of the belt.
        This method is called, if the belt would like to be activated or updated.

        Parameters
        ----------
        now : float
            current time of the system.
        duration : float
            duration of the belt being activated, which being defined by the time set.
        speed : float
            speed of the belt or the action by an agent in RL context.
        overwrite : bool, optional
            To indicate whether the current operation can be overwritten or not.
        """
        if speed > 1.0:
            self.speed = 1.0
        else:
            self.speed = speed
        if self.status == False or overwrite:
            self.t_activated = now
            
            if self.actiontype == "C":
                self.t_end = now + duration
                self.cur_action = self.speed
            elif self.actiontype == "B":
                self.t_end = now + duration * self.action_max
                self.cur_action = duration * self.action_max
            self.status = True
            

## -------------------------------------------------------------------------------------------------
    def calc_mass(self, now):
        """
        This method calculates the transported mass flow by the belt for a time step.

        Parameters
        ----------
        now : float
            current time of the system.

        Returns
        -------
        cur_mass_transport : float
            current transported mass.

        """
        if self.status == True:
            if self.actiontype == "C":
                self.cur_mass_transport = self.mass_coeff*(self.speed*(self.action_max-self.action_min)+self.action_min)
            elif self.actiontype == "B":
               self.cur_mass_transport = self.mass_coeff*self.speed
        return self.cur_mass_transport
    

## -------------------------------------------------------------------------------------------------
    def calc_power(self):
        """
        This method calculates the power consumption of a belt.

        Returns
        -------
        cur_power : float
            current power consumption.

        """
        if self.status == True:
            if self.actiontype == "C":
                self.cur_power  = self.speed*(self.power_max-self.power_min)+self.power_min
            elif self.actiontype == "B":
                self.cur_power = self.power_max
        return self.cur_power / 1000.0
        

## -------------------------------------------------------------------------------------------------
    def update(self, now):
        """
        This method calculates whether a belt must be deactived or not.

        Parameters
        ----------
        now : float
            current time of the system.

        """
        if self.status == True:
            if self.t_end < now:
                self.deactivate()
    

## -------------------------------------------------------------------------------------------------
    def deactivate(self):
        """
        This method is used to deactivate a belt.
        """
        self.status             == False
        self.cur_mass_transport = 0
        self.cur_power          = 0




## -------------------------------------------------------------------------------------------------
## -------------------------------------------------------------------------------------------------
class Reservoir:
    """
    This class serves as a parent class of different types of reservoirs, which provides the main 
    attributes of a buffer in the BGLP environment.

    Parameters
    ----------
    vol_max : float
        maximum volume of a reservoir.
    vol_init_abs : float, optional
        initial volume of a reservoir. The default is 0.
        
    Attributes
    ----------
    reg_r : list of objects
        list of existing reservoirs in the environment.
    idx_r : int
        length of reg_r.
    vol_max : float
        maximum volume of a reservoir.
    vol_init_abs : float
        initial volume of a reservoir.
    vol_cur_abs : float
        current volume of a reservoir.
    vol_cur_rel : float
        current volume of a reservoir in percentage.
    change : float
        volume change of a reservoir in a time step.
    """
    reg_r           = []
    idx_r           = []
    vol_max         = 0
    vol_init_abs    = 0
    vol_cur_abs     = 0
    vol_cur_rel     = 0
    change          = 0

## -------------------------------------------------------------------------------------------------
    def __init__(self, vol_max, vol_init_abs=0):
        self.idx_r          = len(self.reg_r)
        self.reg_r.append(self)
        self.vol_max        = vol_max
        self.vol_init_abs   = vol_init_abs
        self.vol_cur_abs    = vol_init_abs
        self.vol_cur_rel    = self.vol_cur_abs / self.vol_max
        self.change         = 0
        

## -------------------------------------------------------------------------------------------------
    def set_change(self, vol_change):
        """
        This method sets up a volume change of a reservoir.

        Parameters
        ----------
        vol_change : float
            volume change of a reservoir in a time step.
        """
        self.change = vol_change
            

## -------------------------------------------------------------------------------------------------
    def update(self):
        """
        This method calculates the current volume of a reservoir after volume changes are made.
        
        """
        self.vol_cur_abs += self.change
        self.vol_cur_rel = self.vol_cur_abs / self.vol_max




## -------------------------------------------------------------------------------------------------
## -------------------------------------------------------------------------------------------------
class Silo(Reservoir):
    """
    This class inherits Reservoir class and serves as a child class of Reservoir.
    This class represents a type of buffers in the BGLP environment, namely Silo.
    Silos are used to temporary stored the transported materials.
    However, the parameter of each silo can be dissimilar to each other based on their settings.

    Parameters
    ----------
    name : str
        specific name or id of a silo (e.g. Silo_A, etc.).
    vol_max : float
        maximum capacity of a silo.
    vol_cur : float
        current volume of a silo.
    mode : str, optional
        mode of measuring the current volume.
        "abs" means absolute value, "rel" means percentage. The default is "abs".
    
    Attributes
    ----------
    reg_s : list of objects
        list of existing silos.
    idx_s : int
        length of reg_s.
    name : str
        specific name or id of a silo.
    type_ : str
        a short name for a silo, usually 3 capital letters (e.g SIL).
    """
    reg_s   = []
    idx_s   = []
    name    = ""
    type_   = ""
    
## -------------------------------------------------------------------------------------------------
    def __init__(self, name, vol_max, vol_cur = 0, mode = "abs"):
        if mode == "abs":
            vol_cur = vol_cur
        elif mode == "rel":
            vol_cur = vol_max * vol_cur
        Reservoir.__init__(self, vol_max, vol_cur)
        self.name   = name
        self.idx_s  = len(self.reg_s)
        self.reg_s.append(self)
        self.type_  = "SIL"




## -------------------------------------------------------------------------------------------------
## -------------------------------------------------------------------------------------------------
class Hopper (Reservoir):
    """
    This class inherits Reservoir class and serves as a child class of Reservoir.
    This class represents a type of buffers in the BGLP environment, namely Hopper.
    Hoppers are used to temporary stored the transported materials.
    However, the parameter of each hopper can be dissimilar to each other based on their settings.

    Parameters
    ----------
    name : str
        specific name or id of a hopper (e.g. Hop_A, etc.).
    vol_max : float
        maximum capacity of a hopper.
    vol_cur : float
        current volume of a hopper.
    mode : str, optional
        mode of measuring the current volume.
        "abs" means absolute value, "rel" means percentage. The default is "abs".
    
    Attributes
    ----------
    reg_h : list of objects
        list of existing silos.
    idx_h : int
        length of reg_h.
    name : str
        specific name or id of a hopper.
    type_ : str
        a short name for a hopper, usually 3 capital letters (e.g HOP).
    """
    reg_h   = []
    idx_h   = []
    name    = ""
    type_   = ""
    
## -------------------------------------------------------------------------------------------------
    def __init__(self, name, vol_max, vol_cur = 0, mode = "abs"):
        if mode == "abs":
            vol_cur = vol_cur
        elif mode == "rel":
            vol_cur = vol_max * vol_cur
        Reservoir.__init__(self, vol_max, vol_cur)
        self.name   = name
        self.idx_h  = len(self.reg_h)
        self.reg_h.append(self)
        self.type_  = "HOP" 




## -------------------------------------------------------------------------------------------------
## -------------------------------------------------------------------------------------------------
class BGLP (Environment):
    """
    This is the main class of BGLP environment that inherits Environment class from MLPro.
    
    Parameters
    ----------
    p_reward_type : Reward, optional
        rewarding type. The default is Reward.C_TYPE_EVERY_AGENT.
    p_logging : Log, optional
        logging functionalities. The default is Log.C_LOG_ALL.
    t_step : float, optional
        time for each time step (in seconds). The default is 0.5.
    t_set : float, optional
        time set for one horizon (in seconds). The default is 10.0.
    demand : float, optional
        the constant output flow from the system (in L/s). The default is 0.1.
    lr_margin : float, optional
        the learning rate for margin parameter, related to implemented reward function. The default is 1.0.
    lr_demand : float, optional
        the learning rate for production demand, related to implemented reward function. The default is 4.0.
    lr_power : float, optional
        the learning rate for power consumption, related to implemented reward function. The default is 0.0010.
    margin_p : list of floats, optional
        the margin parameter of reservoirs [low, high, multplicator]. The default is [0.2,0.8,4].
    prod_target : float, optional
        the production target for batch operation (in L). The default is 10000.
    prod_scenario : str, optional
        'batch' means batch production scenario and 'continuous' means continuous production scenario. The default is 'continuous'.

    Attributes
    ----------
    C_NAME : str
        name of the environment.
    C_LATENCY : timedelta()
        latency.
    C_INFINITY : np.finfo()
        infinity.
    C_REWARD_TYPE : Reward
        rewarding type.
    sils : list of objects
        list of existing silos.
    hops : list of objects
        list of existing hoppers.
    ress : list of objects
        list of existing reservoirs.
    blts : list of objects
        list of existing belts.
    acts : list of objects
        list of existing actuators.
    vacs : list of objects
        list of existing vacuum pumps.
    con_act_to_res : list of lists of int
        connection between actuators and reservoirs and setup the sequence.
    m_param : list of floats
        the margin parameter of reservoirs [low, high, multplicator].
    _demand : float
        the constant output flow from the system (in L/s).
    t : float
        current time of the system (in seconds).
    t_step : float
        time for each time step (in seconds).
    lr_margin : float
        the learning rate for margin parameter.
    lr_demand : float
        the learning rate for production demand.
    lr_power : float
        the learning rate for power consumption.
    overflow : list of floats
        current overflow.
    power : list of floats
        current power consumption.
    transport : list of floats
        current transported mass flow.
    reward : list of floats
        current rewards.
    levels_init : float
        initial level of reservoirs.
    overflow_t : float
        current overflow for a specific buffer in a specific time.
    demand_t : float
        current demand for a specific buffer in a specific time.
    power_t : float
        current power consumption for a specific actuator in a specific time.
    transport_t : float
        current transported material by a specific actuaor in a specific time.
    margin_t : float
        current margin for a specific buffer in a specific time.
    prod_reached : float
        current production reached in L for batch operation.
    prod_target : float
        the production target for batch operation (in L).
    prod_scenario : str
        'batch' means batch production scenario and 'continuous' means continuous production scenario.
    cycle_limit : int
        the number of cycle limit.
    
    """
    C_NAME              = "BGLP"
    C_LATENCY           = timedelta(0,1,0)    
    C_INFINITY          = np.finfo(np.float32).max
    C_REWARD_TYPE       = Reward.C_TYPE_EVERY_AGENT
    C_CYCLE_LIMIT       = 0
    sils                = []
    hops                = []
    ress                = []
    blts                = []
    vacs                = []
    acts                = []
    con_act_to_res      = []
    m_param             = []
    _demand             = 0
    t                   = 0
    t_step              = 0
    lr_margin           = 0
    lr_demand           = 0
    lr_power            = 0
    overflow            = []
    demand              = []
    power               = []
    transport           = []
    reward              = []
    levels_init         = 0
    reset_levels        = 0
    overflow_t          = 0
    demand_t            = 0
    power_t             = 0
    transport_t         = 0
    margin_t            = 0
    prod_reached        = 0
    prod_target         = 0
    prod_scenario       = 0

## -------------------------------------------------------------------------------------------------
    def __init__(self, p_reward_type=Reward.C_TYPE_EVERY_AGENT, p_logging=Log.C_LOG_ALL,
                 t_step=0.5, t_set=10.0, demand=0.1, lr_margin=1.0, lr_demand=4.0,
                 lr_power=0.0010, margin_p=[0.2,0.8,4], prod_target=10000,
                 prod_scenario='continuous', cycle_limit=0):
        self.num_envs       = 5                                                 # Number of internal sub-environments
        self.reward_type    = p_reward_type
        super().__init__(p_mode=Environment.C_MODE_SIM, p_logging=p_logging)
        
        self.C_SCIREF_TYPE    = self.C_SCIREF_TYPE_ARTICLE
        self.C_SCIREF_AUTHOR  = "Dorothea Schwung, Steve Yuwono, Andreas Schwung, Steven X. Ding"
        self.C_SCIREF_TITLE   = "Decentralized learning of energy optimal production policies using PLC-informed reinforcement learning"
        self.C_SCIREF_JOURNAL = "Computers & Chemical Engineering"
        self.C_SCIREF_YEAR    = "2021"
        self.C_SCIREF_MONTH   = "05"
        self.C_SCIREF_DAY     = "28"
        self.C_SCIREF_VOLUME  = "152"
        self.C_SCIREF_DOI     = "10.1016/j.compchemeng.2021.107382"
        
        self.C_CYCLE_LIMIT  = cycle_limit
        self.t              = 0
        self.t_step         = t_step
        self.t_set          = t_set
        self._demand        = demand
        self.lr_margin      = lr_margin
        self.lr_demand      = lr_demand
        self.lr_power       = lr_power
        self.prod_target    = prod_target
        self.prod_scenario  = prod_scenario
        self.levels_init    = np.ones((6,1))*0.5
        self.sils           = []
        self.hops           = []
        self.ress           = []
        self.blts           = []
        self.vacs           = []
        self.acts           = []
        
        belt_a              = Belt("Belt_A", "C", 40.0, 50.5, 450, 1850, 0.01/60)
        belt_b              = Belt("Belt_B", "B", 0, 26.9, 0, self.t_set, 0.4)
        belt_c              = Belt("Belt_C", "C", 114.828, 370, 450, 1450, 0.01249/60)
        vac_b               = VacuumPump("Vac_B", 0, 305, 0.567, 4.575*int(self.t_set//4.575), [0.464, 0.0332])
        vac_c               = VacuumPump("Vac_C", 0, 456, 0.979, 9.5*int(self.t_set//9.5), [0.3535, 0.0096])
        
        sil_a               = Silo("Silo_A", 17.42, random.uniform(0,1), mode="rel")                
        sil_b               = Silo("Silo_B", 17.42, random.uniform(0,1), mode="rel")
        sil_c               = Silo("Silo_C", 17.42, random.uniform(0,1), mode="rel")
        hop_a               = Hopper("Hopper_A", 9.1, random.uniform(0,1), mode="rel")                
        hop_b               = Hopper("Hopper_B", 9.1, random.uniform(0,1), mode="rel")
        hop_c               = Hopper("Hopper_C", 9.1, random.uniform(0,1), mode="rel")
        
        self.sils.append(sil_a)
        self.sils.append(sil_b)
        self.sils.append(sil_c)
        self.hops.append(hop_a)
        self.hops.append(hop_b)
        self.hops.append(hop_c)
        self.ress.append(sil_a)
        self.ress.append(hop_a)
        self.ress.append(sil_b)
        self.ress.append(hop_b)
        self.ress.append(sil_c)
        self.ress.append(hop_c)
        
        self.blts.append(belt_a)
        self.blts.append(belt_b)
        self.blts.append(belt_c)
        self.vacs.append(vac_b)
        self.vacs.append(vac_c)
        self.acts.append(belt_a)
        self.acts.append(vac_b)
        self.acts.append(belt_b)
        self.acts.append(vac_c)
        self.acts.append(belt_c)
        
        self.margin_p           = margin_p
        self.margin             = np.zeros((len(self.ress),1))
        self.overflow           = np.zeros((len(self.ress),1))
        self.demand             = np.zeros((len(self.ress),1))
        self.power              = np.zeros((len(self.acts),1))
        self.transport          = np.zeros((len(self.acts),1))
        self.overflow_t         = np.zeros((len(self.ress),1))
        self.demand_t           = np.zeros((len(self.ress),1))
        self.power_t            = np.zeros((len(self.acts),1))
        self.transport_t        = np.zeros((len(self.acts),1))
        self.margin_t           = np.zeros((len(self.ress),1))
        self.reward             = np.zeros((len(self.acts),1))
        self.con_res_to_act     = [[-1,0],[0,1],[1,2],[2,3],[3,4],[4,-1]]
        
        self.data_lists         = ["time","overflow","power","demand"]
        self.data_storing       = DataStoring(self.data_lists)
        self.data_frame         = None
        
        self.reset()
            

## -------------------------------------------------------------------------------------------------
    @staticmethod
    def setup_spaces():
        """
        This method is used to setup action and state spaces of the system.

        Returns
        -------
        state_space : ESpace()
            state space of the system.
        action_space : ESpace()
            action space of the system.

        """
        state_space     = ESpace()
        action_space    = ESpace()
        levels_max      = [17.42, 9.10, 17.42, 9.10, 17.42, 9.10]

        state_space.add_dim(Dimension('E-0 LvlSiloA', 'R', 'Env-0 Level of Silo A', '', 'L', 'L',[0, levels_max[0]]))
        state_space.add_dim(Dimension('E-0 LvlHopperA', 'R', 'Env-0 Level of Hopper A', '', 'L', 'L',[0, levels_max[1]]))
        state_space.add_dim(Dimension('E-0 LvlSiloB', 'R', 'Env-0 Level of Silo B', '', 'L', 'L',[0, levels_max[2]]))
        state_space.add_dim(Dimension('E-0 LvlHopperB', 'R', 'Env-0 Level of Hopper B', '', 'L', 'L',[0, levels_max[3]]))
        state_space.add_dim(Dimension('E-0 LvlSiloC', 'R', 'Env-0 Level of Silo C', '', 'L', 'L',[0, levels_max[4]]))
        state_space.add_dim(Dimension('E-0 LvlHopperC', 'R', 'Env-0 Level of Hopper C', '', 'L', 'L',[0, levels_max[5]]))
        
        action_space.add_dim(Dimension('E-0 Act', 'R', 'Env-0 Actuator Control', '', '', '', [0,1]))
        action_space.add_dim(Dimension('E-1 Act', 'R', 'Env-1 Actuator Control', '', '', '', [0,1]))
        action_space.add_dim(Dimension('E-2 Act', 'Z', 'Env-2 Actuator Control', '', '', '', [0,1]))
        action_space.add_dim(Dimension('E-3 Act', 'R', 'Env-3 Actuator Control', '', '', '', [0,1]))
        action_space.add_dim(Dimension('E-4 Act', 'R', 'Env-4 Actuator Control', '', '', '', [0,1]))

        return state_space, action_space


## -------------------------------------------------------------------------------------------------
    def collect_substates(self) -> State:
        """
        This method is called during resetting the environment.

        Returns
        -------
        state : State
            current states.

        """
        state = State(self._state_space)
        sub_state_val = self.calc_state()
        for i in range(len(sub_state_val)):
            state.set_value(state.get_dim_ids()[i], sub_state_val[i])
        return state
    

## -------------------------------------------------------------------------------------------------
    def _reset(self, p_seed=None) -> None:
        """
        This method is used to reset the environment.

        Parameters
        ----------
        p_seed : int
            Seed parameter for an internal random generator.

        """
        np.random.seed(p_seed)
        self.levels_init = np.random.rand(6,1)
        self.reset_levels()
        self.reset_actuators()
        obs                 = self.calc_state()
        self.t              = 0
        self.prod_reached   = 0
        self._state         = self.collect_substates()
        self.get_state().set_success(False)
        self.get_state().set_broken(False)
        if self.data_frame == None:
            self.data_frame = 0
        else:
            self.data_frame += 1
        self.data_storing.add_frame(str(self.data_frame))


## -------------------------------------------------------------------------------------------------
    def _simulate_reaction(self, p_state:State, p_action:Action) -> State:
        """
        This method is used to calculate the next states of the system after a set of actions.

        Parameters
        ----------
        p_state : State
            State.
        p_action : Action
            ACtion.

        Returns
        -------
        _state : State
            current states.

        """
        action          = []
        for agent_id in p_action.get_agent_ids():
            action_elem = p_action.get_elem(agent_id)
            for action_id in action_elem.get_dim_ids():
                action.append(action_elem.get_value(action_id))
        
        self.overflow_t         = np.zeros((len(self.ress),1))
        self.demand_t           = np.zeros((len(self.ress),1))
        self.power_t            = np.zeros((len(self.acts),1))
        self.transport_t        = np.zeros((len(self.acts),1))
        self.margin_t           = np.zeros((len(self.ress),1))
        
        x = 0
        while x < (self.t_set//self.t_step):
            overflow_diff, demand_diff, power_diff, transport_diff, margin_diff = self.get_status(self.t, self._demand)
            self.overflow_t     += overflow_diff
            self.demand_t       += demand_diff
            self.power_t       += power_diff
            self.transport_t    += transport_diff
            self.margin_t       += margin_diff
            self.t              += self.t_step
            x += 1
            
        self.set_actions(action)
        self._state.set_success(False)
        self._state.set_broken(False)
        self._state = self.collect_substates()
        
        self.data_storing.memorize("time",str(self.data_frame),self.t)
        self.data_storing.memorize("overflow",str(self.data_frame), (sum(self.overflow_t[:])/self.t_set).item())
        self.data_storing.memorize("power",str(self.data_frame), (sum(self.power_t[:])/self.t_set).item())
        self.data_storing.memorize("demand",str(self.data_frame), (self.demand_t[-1]/self.t_set).item())

        return self._state


## -------------------------------------------------------------------------------------------------
    def _compute_success(self, p_state:State) -> bool:
        """
        This method computes the success flag. This method can be redefined.

        Parameters
        ----------
        p_state : State
            state.

        Returns
        -------
        bool
            success or not success.

        """        
        if self.prod_scenario == 'continuous':
            return False
        else:
            if self.prod_reached >= self.prod_target:
                self._state.set_terminal(True)
                return True
            else:
                return False
    

## -------------------------------------------------------------------------------------------------
    def _compute_broken(self, p_state:State) -> bool:
        """
        This method computes the broken flag. This method can be redefined.

        Parameters
        ----------
        p_state : State
            state.

        Returns
        -------
        bool
            broken or not.

        """ 
        return False
    

## -------------------------------------------------------------------------------------------------
    def _compute_reward(self, p_state_old:State, p_state_new:State) -> Reward:
        """
        This method calculates the reward for different reward types.

        Parameters
        ----------
        p_state_old : State
            previous state.
        p_state_new : State
            new state.

        Returns
        -------
        reward : Reward
            reward values.

        """
        reward = Reward(self.reward_type)

        if self.reward_type == Reward.C_TYPE_OVERALL:
            r_overall = 0
            r_overall = r_overall + sum(self.calc_reward()).item()
            reward.set_overall_reward(r_overall)
        
        elif self.reward_type == Reward.C_TYPE_EVERY_AGENT:
           for agent_id in self._last_action.get_agent_ids():
               r_reward = self.calc_reward()
               reward.add_agent_reward(agent_id, r_reward[int(agent_id)])
               
        else:
           for agent_id in self._last_action.get_agent_ids():
                agent_action_elem = self._last_action.get_elem(agent_id)
                agent_action_ids = agent_action_elem.get_dim_ids()
                r_agent = 0
                r_reward = self.calc_reward()
                action_idx = 0
                for action_id in agent_action_ids:
                    r_action = r_reward[action_idx]
                    action_idx += 1
                    reward.add_action_reward(agent_id, action_id, r_action)
                    
        return reward

  
## -------------------------------------------------------------------------------------------------
    def calc_mass_flows(self):
        """
        This method calculates the mass flow transported by actuators.
        
        """
        for act_num in range(len(self.acts)):
            self.transport[act_num] = self.acts[act_num].calc_mass(self.t)*self.t_step


## -------------------------------------------------------------------------------------------------
    def calc_power(self):
        """
        This method calculates the power consumptions per actuator.
        
        """
        for act_num in range(len(self.acts)):
            self.power[act_num] = self.acts[act_num].calc_power()*self.t_step
                

## -------------------------------------------------------------------------------------------------
    def calc_margin(self):   
        """
        This method calculates margin level for every reservoir.
        
        """             
        for i in range(len(self.ress)):
            vol_rel = self.ress[i].vol_cur_rel
            if vol_rel < self.margin_p[0]:
                self.margin[i] = (0-self.margin_p[2])/(self.margin_p[0])*(vol_rel-self.margin_p[0])*self.t_step
            elif vol_rel > self.margin_p[1]:
                self.margin[i] = self.margin_p[2]/(1-self.margin_p[1])*(vol_rel-self.margin_p[1])*self.t_step
            else:
                self.margin[i] = 0.0
                

## -------------------------------------------------------------------------------------------------
    def set_volume_changes(self, demandval): 
        """
        This method sets up volume changes for every reservoir.
        
        """
        for resnum in range(len(self.ress)):
            res = self.ress[resnum]
            current_vol = res.vol_cur_abs
            if resnum == 0:
                ins = 0
                outs = self.transport[self.con_res_to_act[resnum][1]]
            elif resnum == len(self.ress)-1:
                ins = self.transport[self.con_res_to_act[resnum][0]]
                outs = demandval*self.t_step
            else:
                ins = self.transport[self.con_res_to_act[resnum][0]]
                outs = self.transport[self.con_res_to_act[resnum][1]]
                
            if ins > self.ress[resnum-1].vol_cur_abs and resnum != 0:
                ins = self.ress[resnum-1].vol_cur_abs
            demand = current_vol+ins-outs
            if outs > current_vol:
                outs = current_vol
            if demand > 0:
                demand = 0
            overflow = current_vol+ins-outs-res.vol_max
            if overflow < 0:
                overflow = 0
            self.overflow[resnum] = overflow
            self.demand[resnum] = demand
            res.set_change(ins-outs-overflow)
            if resnum == len(self.ress)-1:
                self.prod_reached += outs


## -------------------------------------------------------------------------------------------------
    def update_levels(self):
        """
        This method updates the current level of reservoirs.
        
        """
        for resnum in range(len(self.ress)):
            res = self.ress[resnum]
            res.update()
            if resnum == 0:
                if res.vol_cur_rel <= self.margin_p[0]:
                    res.vol_cur_abs = self.levels_init[resnum]*res.vol_max
                    res.vol_cur_rel = self.levels_init[resnum]


## -------------------------------------------------------------------------------------------------
    def reset_levels(self):
        """
        This method resets reservoirs.
        
        """
        for resnum in range(len(self.ress)):
            res = self.ress[resnum]
            res.vol_cur_abs = self.levels_init[resnum]*res.vol_max
            res.vol_cur_rel = self.levels_init[resnum]
    

## -------------------------------------------------------------------------------------------------
    def reset_actuators(self):
        """
        This method resets actuators.
        
        """
        for act in self.acts:
            act.deactivate()


## -------------------------------------------------------------------------------------------------
    def update_actuators(self):
        """
        This method updates actuators.
        
        """
        for act in self.acts:
            act.update(self.t)
                

## -------------------------------------------------------------------------------------------------
    def update(self, demandval):
        """
        This method sets up volume changes, updates reservoirs' level, and updates actuators.
        
        Parameters
        ----------
        demandval : float
            production outflow target in L/s.
            
        """
        self.set_volume_changes(demandval)
        self.update_levels()
        self.update_actuators()
        

## -------------------------------------------------------------------------------------------------
    def get_status(self, t, demandval):
        """
        This method calculates overflow, demand, power, transport, and margin.
        This function will be called every time step.
        
        Parameters
        ----------
        t : float
            current time in sec.
        demandval : float
            production outflow target in L/s.

        Returns
        -------
        overflow : list of floats
            overflow levels.
        demand : list of floats
            demand fulfilled.
        power : list of floats
            power consumptions.
        transport : list of floats
            transported materials.
        margin : list of floats
            margin levels.

        """
        self.t = t       
        self.calc_mass_flows()
        self.calc_power()
        self.calc_margin()
        self.update(demandval)        
        return self.overflow, self.demand, self.power, self.transport, self.margin
            

## -------------------------------------------------------------------------------------------------
    def set_actions(self, actions):   
        """
        This method sets up actions for actuators. This function will be called every time set.
        
        """
        t_set = self.t_set-2*self.t_step  
        for actnum in range(len(self.acts)):
            act = self.acts[actnum]
            if act.type_ == "VAC":
                act.start_t(self.t, actions[actnum])
            elif act.type_ == "BLT":
                if act.actiontype == "B":
                    if actions[actnum] >= 0.5:
                        actions[actnum] = 1
                    else:
                        actions[actnum] = 0
                act.start_t(self.t, t_set, actions[actnum])
    

## -------------------------------------------------------------------------------------------------
    def calc_state(self):
        """
        This method obtains current levels of reservoirs.

        Returns
        -------
        levels : list of floats
            level of each reservoir.
        
        """
        levels = np.zeros((len(self.ress),1))
        for resnum in range(len(self.ress)):
            res = self.ress[resnum]
            levels[resnum] = res.vol_cur_rel
        return levels
            

## -------------------------------------------------------------------------------------------------
    def calc_reward(self):
        """
        This method calculates the reward. This method can be redifined!
        The current reward function is suitable for continuous operation and scalar reward for individual agents.

        Returns
        -------
        reward : list of floats
            reward for each agent.
        """
        for actnum in range(len(self.acts)):
            acts = self.acts[actnum]
            self.reward[actnum] = 1/(1+self.lr_margin*self.margin_t[actnum])+1/(1+self.lr_power*self.power_t[actnum]/(acts.power_max/1000.0))
            if actnum == len(self.acts)-1:
                self.reward[actnum] += 1/(1-self.lr_demand*self.demand_t[-1])
            else:
                self.reward[actnum] += 1/(1+self.lr_margin*self.margin_t[actnum+1])
        return self.reward[:]


## -------------------------------------------------------------------------------------------------
    def init_plot(self, p_figure=None):
        pass


## -------------------------------------------------------------------------------------------------
    def update_plot(self):
        pass